# -*- coding: utf-8 -*-
#
#  Copyright (C) 2017, 2020, 2023 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import os.path as osp

import pyficache

from trepan.lib.stack import frame2file
from trepan.processor.parse.semantics import Location

INVALID_LOCATION = None


def resolve_location(proc, location):
    """Expand fields in Location namedtuple. If:
    '.':  get fields from stack
    function/module: get fields from evaluation/introspection
    location file and line number: use that
    """
    curframe = proc.curframe
    offset = None
    if location == ".":
        if not curframe:
            proc.errmsg("Don't have a stack to get location from")
            return INVALID_LOCATION
        filename = frame2file(proc.core, curframe, canonic=False)
        lineno = inspect.getlineno(curframe)
        offset = curframe.f_lasti
        return Location(filename, lineno, False, None, offset)

    assert isinstance(location, Location)
    is_address = location.is_address
    if proc.curframe:
        g = curframe.f_globals
        locals_dict = curframe.f_locals
    else:
        g = globals()
        locals_dict = locals()
        pass
    if location.method:
        # Validate arguments that can't be done in parsing
        filename = lineno = None
        msg = "Object %s is not known yet as a function, " % location.method
        try:
            modfunc = eval(location.method, g, locals_dict)
        except Exception:
            proc.errmsg(msg)
            return INVALID_LOCATION

        try:
            # Check if the converted string is a function or instance
            # method.  We don't want to test on attributes and not use
            # `inspect.isfunction()` so that we can accommodate
            # trepan-xpy() which has it's own type of compatible
            # Function, that would fail an `inspect.isfunction()`
            # test.
            if hasattr(modfunc, "__code__") or hasattr(modfunc, "im_func"):
                offset = -1
            else:
                proc.errmsg(msg)
                return INVALID_LOCATION
        except Exception:
            proc.errmsg(msg)
            return INVALID_LOCATION
        filename = proc.core.canonic(modfunc.__code__.co_filename)
        # FIXME: we may want to check lineno and
        # respect that in the future
        lineno = modfunc.__code__.co_firstlineno
    elif location.path:
        filename = proc.core.canonic(location.path)
        lineno = location.line_number
        modfunc = None
        msg = "%s is not known as a file" % location.path
        if not osp.isfile(filename):
            # See if argument is a module
            try:
                modfunc = eval(location.path, g, locals_dict)
            except Exception:
                msg = (
                    "Don't see '%s' as a existing file or as an module" % location.path
                )
                proc.errmsg(msg)
                return INVALID_LOCATION
            pass
            is_address = location.is_address
            if inspect.ismodule(modfunc):
                if hasattr(modfunc, "__file__"):
                    filename = pyficache.resolve_name_to_path(modfunc.__file__)
                    filename = proc.core.canonic(filename)
                    if not lineno:
                        # use first line of module file
                        lineno = 1
                        offset = 0
                        is_address = False
                    return Location(filename, lineno, is_address, modfunc, offset)
                else:
                    msg = (
                        "module '%s' doesn't have a file associated with it"
                        % location.path
                    )

            proc.errmsg(msg)
            return INVALID_LOCATION
        maxline = pyficache.maxline(filename)
        if maxline and lineno > maxline:
            # NB: we use the gdb wording here
            proc.errmsg(
                "Line number %d out of range; %s has %d lines."
                % (lineno, filename, maxline)
            )
            return INVALID_LOCATION
        offset = location.offset
        if offset is None:
            lineinfo = pyficache.code_line_info(filename, lineno)
            if lineinfo:
                offset = lineinfo[0].offsets[0]
                modfunc = lineinfo[0].name
            else:
                print("No offset found for %s %s" % (filename, lineno))

    elif location.line_number:
        filename = frame2file(proc.core, curframe, canonic=False)
        lineno = location.line_number
        is_address = location.is_address
        modfunc = None
        if offset is None:
            lineinfo = pyficache.code_line_info(filename, lineno, include_offsets=True)
            if lineinfo:
                offset = lineinfo[0].offsets[0]
                modfunc = lineinfo[0].name
    elif location.offset is not None:
        filename = frame2file(proc.core, curframe, canonic=False)
        is_address = True
        lineno = None
        modfunc = None
        offset = location.offset
    return Location(filename, lineno, is_address, modfunc, offset)


def resolve_address_location(proc, location):
    """Expand fields in Location namedtuple. If:
    '.':  get fields from stack
    function/module: get fields from evaluation/introspection
    location file and line number: use that
    """
    curframe = proc.curframe
    if location == ".":
        filename = frame2file(proc.core, curframe, canonic=False)
        offset = curframe.f_lasti
        is_address = True
        return Location(filename, offset, False, None, offset)

    assert isinstance(location, Location)
    is_address = True
    if proc.curframe:
        g = curframe.f_globals
        locals_dict = curframe.f_locals
    else:
        g = globals()
        locals_dict = locals()
        pass
    if location.method:
        # Validate arguments that can't be done in parsing
        filename = offset = None
        msg = "Object %s is not known yet as a function, " % location.method
        try:
            modfunc = eval(location.method, g, locals_dict)
        except Exception:
            proc.errmsg(msg)
            return INVALID_LOCATION

        try:
            # Check if the converted string is a function or instance method
            if inspect.isfunction(modfunc) or hasattr(modfunc, "im_func"):
                pass
            else:
                proc.errmsg(msg)
                return INVALID_LOCATION
        except Exception:
            proc.errmsg(msg)
            return INVALID_LOCATION
        filename = proc.core.canonic(modfunc.func_code.co_filename)
        # FIXME: we may want to check offset and
        # respect that in the future
        offset = 0
    elif location.path:
        filename = proc.core.canonic(location.path)
        offset = location.line_number
        is_address = location.is_address
        modfunc = None
        msg = "%s is not known as a file" % location.path
        if not osp.isfile(filename):
            # See if argument is a module
            try:
                modfunc = eval(location.path, g, locals_dict)
            except Exception:
                msg = (
                    "Don't see '%s' as a existing file or as an module" % location.path
                )
                proc.errmsg(msg)
                return INVALID_LOCATION
            pass
            is_address = location.is_address
            if inspect.ismodule(modfunc):
                if hasattr(modfunc, "__file__"):
                    filename = pyficache.resolve_name_to_path(modfunc.__file__)
                    filename = proc.core.canonic(filename)
                    if not offset:
                        # use first offset of module file
                        offset = 0
                        is_address = True
                    return Location(filename, offset, is_address, modfunc, offset)
                else:
                    msg = (
                        "module '%s' doesn't have a file associated with it"
                        % location.path
                    )

            proc.errmsg(msg)
            return INVALID_LOCATION
        maxline = pyficache.maxline(filename)
        if maxline and offset > maxline:
            # NB: we use the gdb wording here
            proc.errmsg(
                "Line number %d out of range; %s has %d lines."
                % (offset, filename, maxline)
            )
            return INVALID_LOCATION
    elif location.line_number is not None:
        filename = frame2file(proc.core, curframe, canonic=False)
        offset = location.line_number
        is_address = location.is_address
        modfunc = proc.list_object
    else:
        proc.errmsg(
            f"Location {location} doesn't have enough information for a location."
        )
        return INVALID_LOCATION

    return Location(filename, offset, is_address, modfunc, offset)



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.lib import complete as Mcomplete
from trepan.processor import frame as Mframe

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class InfoCode(Mbase_subcmd.DebuggerSubcommand):
    """**info code** [ *frame-number* | *python code object* ]

    Show the detailed information for the Python code object in
    *frame-number* or the current frame if *frame-number* is not specified.

    Specific information includes:

    * the number of arguments (not including * or ** args)

    * constants used in the bytecode

    * name of file in which this code object was created

    * number of first line in Python source code

    * name with which this code object was defined

    See also:
    ---------

    `info frame`, `info locals`, `info file`"""

    min_abbrev = 2
    max_args = 2
    need_stack = True
    short_help = """Show detailed info about the Python code object"""

    def complete(self, prefix):
        proc_obj = self.proc
        low, high = Mframe.frame_low_high(proc_obj, None)
        ary = [str(low + i) for i in range(high - low + 1)]
        # FIXME: add in Thread names
        return Mcomplete.complete_token(ary, prefix)

    def run(self, args):
        proc = self.proc
        frame = proc.curframe
        if not frame:
            self.errmsg("No frame selected.")
            return False

        frame_num = None
        frame_num = None
        if len(args) == 1:
            try:
                frame_num = int(args[0])
                i_stack = len(proc.stack)
                if frame_num < -i_stack or frame_num > i_stack - 1:
                    self.errmsg(
                        (
                            "a frame number number has to be in the range %d to %d."
                            + " Got: %d (%s)."
                        )
                        % (-i_stack, i_stack - 1, frame_num, args[0])
                    )
                    return False
                frame = proc.stack[frame_num][0]
                code = frame.f_code
            except Exception:
                try:
                    code = eval(args[0], frame.f_globals, frame.f_locals)
                except Exception:
                    self.errmsg(
                        "%s is not a evaluable as a code object or frame number."
                        % args[0]
                    )
                    return False
                # if not inspect.iscode(code):
                #     self.errmsg('%s is not a code object' % code)
                pass

        else:
            frame_num = proc.curindex
            code = frame.f_code

        mess = "Code for Frame %d" % frame_num if frame_num is not None else "Code Info"
        self.section(mess)
        self.msg("  name: %s" % code.co_name)
        self.msg("  number of arguments: %d" % code.co_argcount)
        self.msg("  number of locals: %d" % code.co_nlocals)
        self.msg("  maximum stack size %s" % code.co_stacksize)
        self.msg("  first line number: %s" % code.co_firstlineno)
        self.msg("  is%s optimized" % ("" if (code.co_flags & 1) == 1 else " not"))
        self.msg(
            "  new local namespace? %s" % ("yes" if (code.co_flags & 2) == 1 else " no")
        )
        self.msg("  has%s *args" % ("" if (code.co_flags & 4) == 1 else " no"))
        self.msg("  has%s **args" % ("" if (code.co_flags & 8) == 1 else " no"))
        maxwidth = self.settings["width"] // 2
        for name, field in [
            ("Constants", "co_consts"),
            ("Variable names", "co_varnames"),
            ("Local Variables", "co_names"),
        ]:
            vals = [proc._saferepr(x, maxwidth) for x in getattr(code, field)]
            if vals:
                self.section(name)
                m = self.columnize_commands(vals)
                self.msg_nocr(m)

        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import info as Minfo, mock

    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoCode(i)
    import inspect

    cp.curframe = inspect.currentframe()
    sub.run([])



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2007-2010, 2015, 2020, 2023 Rocky Bernstein

#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import types


def print_dict(s, obj, title):
    if hasattr(obj, "__dict__"):
        obj = obj.__dict__
        pass
    if isinstance(obj, dict):
        s += "\n%s:\n" % title
        keys = list(obj.keys())
        keys.sort()
        for key in keys:
            s += "  %s:\t%s\n" % (repr(key), obj[key])
            pass
        pass
    return s


def print_argspec(obj, obj_name):
    """A slightly decorated version of inspect.format_argspec"""
    try:
        return obj_name + inspect.formatargspec(*inspect.getargspec(obj))
    except Exception:
        return None
    return  # Not reached


def print_obj(arg, frame, format=None, short=False):
    """Return a string representation of an object"""
    try:
        if not frame:
            # ?? Should we have set up a dummy globals
            # to have persistence?
            obj = eval(arg, None, None)
        else:
            obj = eval(arg, frame.f_globals, frame.f_locals)
            pass
    except Exception:
        return 'No symbol "' + arg + '" in current context.'
    # format and print
    what = arg
    if format:
        what = format + " " + arg
        obj = printf(obj, format)
    s = "%s = %s" % (what, obj)
    if not short:
        s += "\ntype = %s" % type(obj)
        if callable(obj):
            argspec = print_argspec(obj, arg)
            if argspec:
                s += ":\n\t"
                if inspect.isclass(obj):
                    s += "Class constructor information:\n\t"
                    obj = obj.__init__
                elif isinstance(obj, types.InstanceType):
                    obj = obj.__call__
                    pass
                s += argspec
            pass

        # Try to list the members of a class.
        # Not sure if this is correct or the
        # best way to do.
        s = print_dict(s, obj, "object variables")
        if hasattr(obj, "__class__"):
            s = print_dict(s, obj.__class__, "class variables")
            pass
    return s


pconvert = {"c": chr, "x": hex, "o": oct, "f": float, "s": str}
twos = (
    "0000",
    "0001",
    "0010",
    "0011",
    "0100",
    "0101",
    "0110",
    "0111",
    "1000",
    "1001",
    "1010",
    "1011",
    "1100",
    "1101",
    "1110",
    "1111",
)


def printf(val, fmt):
    global pconvert, twos
    if not fmt:
        fmt = " "  # not 't' nor in pconvert
    # Strip leading '/'
    if fmt[0] == "/":
        fmt = fmt[1:]
    f = fmt[0]
    if f in pconvert.keys():
        try:
            return pconvert[f](val)
        except Exception:
            return str(val)
    # binary (t is from 'twos')
    if f == "t":
        try:
            res = ""
            while val:
                res = twos[val & 0xF] + res
                val = val >> 4
            return res
        except Exception:
            return str(val)
    return str(val)


if __name__ == "__main__":
    print(print_dict("", globals(), "my globals"))
    print("-" * 40)
    print(print_obj("print_obj", None))
    print("-" * 30)
    print(print_obj("Exception", None))
    print("-" * 30)
    print(print_argspec("Exception", None))

    class Foo:
        def __init__(self, bar=None):
            pass

        pass

    print(print_obj("Foo.__init__", None))
    print("-" * 30)
    print(print_argspec(Foo.__init__, "__init__"))
    assert printf(31, "/o") == "037"
    assert printf(31, "/t") == "00011111"
    assert printf(33, "/c") == "!"
    assert printf(33, "/x") == "0x21"



# -*- coding: utf-8 -*-
#   Copyright (C) 2013, 2020, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Command completion routines."""

import re


def complete_token(complete_ary, prefix):
    return sorted([cmd for cmd in complete_ary if cmd.startswith(prefix)])


def complete_token_with_next(complete_hash, prefix, cmd_prefix=""):
    result = []
    for cmd_name in list(complete_hash.keys()):
        if cmd_name.startswith(cmd_prefix + prefix):
            result.append([cmd_name[len(cmd_prefix) :], complete_hash[cmd_name]])
            pass
        pass
    pass
    return sorted(result, key=lambda pair: pair[0])


def complete_token_filtered_with_next(aliases, prefix, expanded, commands):
    """Find all starting matches in dictionary *aliases* that start
    with *prefix*, but filter out any matches already in
    *expanded*."""

    complete_ary = list(aliases.keys())
    expanded_ary = list(expanded.keys())
    # result = [cmd for cmd in
    #             complete_ary if cmd.startswith(prefix) and not (
    #                 cmd in aliases and
    #                 0 == len(set(expanded_ary) - set([aliases[cmd]])))]
    result = []
    for cmd in complete_ary:
        if cmd.startswith(prefix):
            if cmd in aliases and (0 == len(set(expanded_ary) - set([aliases[cmd]]))):
                result.append([cmd, aliases[cmd]])
            pass
        pass
    pass
    return sorted(result, key=lambda pair: pair[0])


def complete_token_filtered(aliases, prefix, expanded):
    """Find all starting matches in dictionary *aliases* that start
    with *prefix*, but filter out any matches already in *expanded*"""

    complete_ary = aliases.keys()
    return [cmd for cmd in complete_ary if cmd.startswith(prefix)]


def complete_brkpts(bpmgr, prefix):
    return complete_token(sorted(bpmgr.bpnumbers()), prefix)


def next_token(str, start_pos):
    """Find the next token in str string from start_pos, we return
    the token and the next blank position after the token or
    str.size if this is the last token. Tokens are delimited by
    white space."""
    look_at = str[start_pos:]
    match = re.search(r"\S", look_at)
    if match:
        pos = match.start()
    else:
        pos = 0
        pass
    next_nonblank_pos = start_pos + pos
    next_match = re.search(r"\s", str[next_nonblank_pos:])
    if next_match:
        next_blank_pos = next_nonblank_pos + next_match.start()
    else:
        next_blank_pos = len(str)
        pass
    return [next_blank_pos, str[next_nonblank_pos : next_blank_pos + 1].rstrip()]


if __name__ == "__main__":
    print(next_token("ab cd ef", 0))
    print(next_token("ab cd ef", 2))
    print(complete_token(("-1", "0"), ""))

    #   0         1
    #   0123456789012345678
    x = "  now is  the  time"
    for pos in [0, 2, 5, 8, 9, 13, 19]:
        print(next_token(x, pos))
        pass
    pass

    print(complete_token(["ba", "aa", "ab"], "a"))
    print(complete_token(["cond", "condition", "continue"], "cond"))
    h = {"ab": 1, "aac": 2, "aa": 3, "b": 4}
    print(complete_token(h.keys(), "a"))
    print(complete_token_with_next(h, "a"))



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
#   Copyright (C) 2013-2015, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import os
import sys
from optparse import OptionParser

from trepan import api as Mapi, clifns as Mclifns, debugger as Mdebugger
from trepan.inout import output as Moutput
from trepan.lib import file as Mfile


def default_configfile(base_filename):
    """Return fully expanded configuration filename location for
    base_filename. python2 and  python3 debuggers share the same
    directory: ~/.config/trepan.py
    """
    file_dir = os.path.join(os.environ.get("HOME", "~"), ".config", "trepanpy")
    file_dir = Mclifns.path_expanduser_abs(file_dir)

    if not os.path.isdir(file_dir):
        os.makedirs(file_dir, mode=0o755)
    return os.path.join(file_dir, base_filename)


def add_startup_file(dbg_initfiles):
    """Read debugger startup file(s): both python code and
    debugger profile to dbg_initfiles."""

    startup_python_file = default_configfile("profile.py")

    if Mfile.readable(startup_python_file):
        with codecs.open(startup_python_file, "r", encoding="utf8") as fp:
            exec(fp.read())

    startup_trepan_file = default_configfile("profile")
    if Mfile.readable(startup_trepan_file):
        if startup_trepan_file not in dbg_initfiles:
            dbg_initfiles.append(startup_trepan_file)
        pass
    return


def process_options(debugger_name, pkg_version, sys_argv, option_list=None):
    """Handle debugger options. Set `option_list' if you are writing
    another main program and want to extend the existing set of debugger
    options.

    The options dictionary from optparser is returned. sys_argv is
    also updated."""
    usage_str = """%prog [debugger-options] [python-script [script-options...]]

    Runs the extended python debugger"""

    # serverChoices = ('TCP','FIFO', None)

    optparser = OptionParser(
        usage=usage_str,
        option_list=option_list,
        version="%%prog version %s" % pkg_version,
    )

    optparser.add_option(
        "-X",
        "--trace",
        dest="linetrace",
        action="store_true",
        default=False,
        help="Show lines before executing them. " "This option also sets --batch",
    )
    optparser.add_option(
        "-F",
        "--fntrace",
        dest="fntrace",
        action="store_true",
        default=False,
        help="Show functions before executing them. " "This option also sets --batch",
    )
    optparser.add_option(
        "--basename",
        dest="basename",
        action="store_true",
        default=False,
        help="Filenames strip off basename, " "(e.g. for regression tests)",
    )
    #     optparser.add_option("--batch", dest="noninteractive",
    #                          action="store_true", default=False,
    #                          help="Don't run interactive commands shell on "+
    #                          "stops.")
    optparser.add_option(
        "--client",
        dest="client",
        action="store_true",
        help="Connect to an existing debugger process "
        "started with the --server option. "
        "See options for client.",
    )
    optparser.add_option(
        "-x",
        "--command",
        dest="command",
        action="store",
        type="string",
        metavar="FILE",
        help="Execute commands from FILE.",
    )
    optparser.add_option(
        "--cd",
        dest="cd",
        action="store",
        type="string",
        metavar="DIR",
        help="Change current directory to DIR.",
    )
    optparser.add_option(
        "--confirm",
        dest="confirm",
        action="store_true",
        default=True,
        help="Confirm potentially dangerous operations",
    )
    optparser.add_option(
        "--dbg_trepan",
        dest="dbg_trepan",
        action="store_true",
        default=False,
        help="Debug the debugger",
    )
    optparser.add_option(
        "--different",
        dest="different",
        action="store_true",
        default=True,
        help="Consecutive stops should have " "different positions",
    )
    #     optparser.add_option("--error", dest="errors", metavar='FILE',
    #                          action="store", type='string',
    #                          help="Write debugger's error output "
    #                          + "(stderr) to FILE")
    optparser.add_option(
        "-e",
        "--exec",
        dest="execute",
        type="string",
        help="list of debugger commands to " + "execute. Separate the commands with ;;",
    )

    optparser.add_option(
        "-H",
        "--host",
        dest="host",
        default="127.0.0.1",
        action="store",
        type="string",
        metavar="IP-OR-HOST",
        help="connect IP or host name. " "Only valid if --client option given.",
    )

    optparser.add_option(
        "--highlight",
        dest="highlight",
        action="store",
        type="string",
        metavar="{light|dark|plain}",
        default="light",
        help="Use syntax and terminal highlight output. " "'plain' is no highlight",
    )

    optparser.add_option(
        "--private",
        dest="private",
        action="store_true",
        default=False,
        help="Don't register this as a global debugger",
    )

    optparser.add_option(
        "--main",
        dest="main",
        action="store_true",
        default=True,
        help="First stop should be in __main__",
    )
    optparser.add_option(
        "--no-main",
        dest="main",
        action="store_false",
        default=True,
        help="First stop should be in __main__",
    )
    optparser.add_option(
        "--post-mortem",
        dest="post_mortem",
        action="store_true",
        default=True,
        help=("Enter debugger on an uncaught (fatal) " "exception"),
    )

    optparser.add_option(
        "--no-post-mortem",
        dest="post_mortem",
        action="store_false",
        default=True,
        help=("Don't enter debugger on an uncaught (fatal) " "exception"),
    )

    optparser.add_option(
        "-n",
        "--nx",
        dest="noexecute",
        action="store_true",
        default=False,
        help=("Don't execute commands found in any " "initialization files"),
    )

    optparser.add_option(
        "-o",
        "--output",
        dest="output",
        metavar="FILE",
        action="store",
        type="string",
        help=("Write debugger's output (stdout) " "to FILE"),
    )
    optparser.add_option(
        "-P",
        "--port",
        dest="port",
        default=1027,
        action="store",
        type="int",
        help="Use TCP port number NUMBER for " "out-of-process connections.",
    )

    optparser.add_option(
        "--server",
        dest="server",
        action="store_true",
        help="Out-of-process server connection mode",
    )

    # optparser.add_option("--style", dest="style",
    #                      action="store", type='string',
    #                      metavar='*pygments-style*',
    #                      default=None,
    #                      help=("Pygments style; 'none' "
    #                            "uses 8-color rather than 256-color terminal"))

    optparser.add_option(
        "--sigcheck",
        dest="sigcheck",
        action="store_true",
        default=False,
        help="Set to watch for signal handler changes",
    )
    optparser.add_option(
        "-t",
        "--target",
        dest="target",
        help=(
            "Specify a target to connect to. Arguments"
            " should be of form, 'protocol address'."
        ),
    ),
    optparser.add_option(
        "--from_ipython",
        dest="from_ipython",
        action="store_true",
        default=False,
        help="Called from inside ipython",
    )

    # annotate option produces annotations, used in trepan.el for a
    # better emacs integration. Annotations are similar in purpose to
    # those of GDB (see that manual for a description), although the
    # syntax is different.  they have the following format:
    #
    # ^Z^Zannotation-name
    # <arbitrary text>
    # ^Z^Z
    #
    # where ^Z is the ctrl-Z character, and "annotname" is the name of the
    # annotation. A line with only two ^Z ends the annotation (no nesting
    # allowed). See trepan.el for the usage
    optparser.add_option(
        "--annotate", default=0, type="int", help="Use annotations to work inside emacs"
    )

    # Set up to stop on the first non-option because that's the name
    # of the script to be debugged on arguments following that are
    # that scripts options that should be left untouched.  We would
    # not want to interpret and option for the script, e.g. --help, as
    # one one of our own, e.g. --help.

    optparser.disable_interspersed_args()

    sys.argv = list(sys_argv)
    # FIXME: why does this mess up integration tests?
    # (opts, sys.argv) = optparser.parse_args(sys_argv)
    (opts, sys.argv) = optparser.parse_args()
    dbg_opts = {"from_ipython": opts.from_ipython}

    # Handle debugger startup command files: --nx (-n) and --command.
    dbg_initfiles = []
    if not opts.noexecute:
        add_startup_file(dbg_initfiles)

    # As per gdb, first we execute user initialization files and then
    # we execute any file specified via --command.
    if opts.command:
        dbg_initfiles.append(opts.command)
        pass

    dbg_opts["proc_opts"] = {"initfile_list": dbg_initfiles}

    if opts.cd:
        os.chdir(opts.cd)
        pass

    if opts.output:
        try:
            dbg_opts["output"] = Moutput.DebuggerUserOutput(opts.output)
        except IOError:
            _, xxx_todo_changeme, _ = sys.exc_info()
            (errno, strerror) = xxx_todo_changeme.args
            print("I/O in opening debugger output file %s" % opts.output)
            print("error(%s): %s" % (errno, strerror))
        except Exception:
            print("Unexpected error in opening debugger output file %s" % opts.output)
            print(sys.exc_info()[0])
            sys.exit(2)
            pass
        pass

    return opts, dbg_opts, sys.argv


def _postprocess_options(dbg, opts):
    """Handle options (`opts') that feed into the debugger (`dbg')"""
    # Set dbg.settings['printset']
    print_events = []
    if opts.fntrace:
        print_events = ["c_call", "c_return", "call", "return"]
    if opts.linetrace:
        print_events += ["line"]
    if len(print_events):
        dbg.settings["printset"] = frozenset(print_events)
        pass

    for setting in ("annotate", "basename", "different"):
        dbg.settings[setting] = getattr(opts, setting)
        pass

    if getattr(opts, "highlight"):
        dbg.settings["highlight"] = opts.highlight
    else:
        dbg.settings["highlight"] = "plain"

    # if getattr(opts, 'style') and opts.style != 'none':
    #     dbg.settings['style'] = opts.style
    # else:
    #     dbg.settings['style'] = None
    dbg.settings["style"] = None

    # Normally we want to set Mdebugger.debugger_obj so that one can
    # put trepan.debugger breakpoints in a program and not have more
    # than one debugger running. More than one debugger may confuse
    # users, e.g. set different might stop at the same line once for
    # each debugger.
    if not opts.private:
        Mdebugger.debugger_obj = dbg
        pass

    #     if opts.errors:
    #         try:
    #             dbg.stderr = open(opts.errors, 'w')
    #         except IOError, (errno, strerror):
    #             print "I/O in opening debugger output file %s" % opts.errors
    #             print "error(%s): %s" % (errno, strerror)
    #         except ValueError:
    #             print "Could not convert data to an integer."
    #         except Exception:
    #             print "Unexpected error in opening debugger output "
    #                   "file %s" % opts.errors
    #             print sys.exc_info()[0]
    #             sys.exit(2)

    #     if opts.execute:
    #         dbg.cmdqueue = list(opts.execute.split(';;'))

    if opts.post_mortem:
        Mapi.debugger_on_post_mortem()
        pass
    return


# Demo it
if __name__ == "__main__":
    import pprint

    def doit(prog, version, arg_str):
        print("options '%s'" % arg_str)
        args = arg_str.split()
        opts, dbg_opts, sys_argv = process_options("testing", version, args)
        pp.pprint(vars(opts))
        print("")
        return

    pp = pprint.PrettyPrinter(indent=4)
    doit("testing", "1.1", "")
    doit("testing", "1.2", "foo bar")
    doit("testing", "1.3", "--server")
    doit("testing", "1.3", "--command %s bar baz" % __file__)
    doit("testing", "1.4", "--server --client")
    doit("testing", "1.5", "--style=emacs")
    doit("testing", "1.6", "--help")  # exits, so must be last
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2013-2014 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Call-frame-oriented helper function for Processor. Put here so we
# can use this in a couple of processors.

from trepan.processor import cmdfns as Mcmdfns
from trepan.lib import complete as Mcomplete


def frame_low_high(proc_obj, direction):
    stack_size = len(proc_obj.stack)  # - hide_level
    if direction is None:
        return [-stack_size, stack_size - 1]
    else:
        frame_index = proc_obj.curindex
        low, high = [
            frame_index * -direction,
            (stack_size - frame_index - 1) * direction,
        ]
        if direction < 0:
            low, high = [high, low]
        return (low, high)
    return


def frame_complete(proc_obj, prefix, direction):
    low, high = frame_low_high(proc_obj, direction)
    ary = [str(low + i) for i in range(high - low + 1)]
    return Mcomplete.complete_token(ary, prefix)


def frame_num(proc_obj, pos):
    return len(proc_obj.stack) - pos - 1


def adjust_frame(proc_obj, name, pos, absolute_pos):
    """Adjust stack frame by pos positions. If absolute_pos then
    pos is an absolute number. Otherwise it is a relative number.

    A negative number indexes from the other end."""
    if not proc_obj.curframe:
        proc_obj.errmsg("No stack.")
        return

    # Below we remove any negativity. At the end, pos will be
    # the new value of proc_obj.curindex.
    if absolute_pos:
        if pos >= 0:
            pos = frame_num(proc_obj, pos)
        else:
            pos = -pos - 1
            pass
    else:
        pos += proc_obj.curindex
        pass

    if pos < 0:
        proc_obj.errmsg("Adjusting would put us beyond the oldest frame.")
        return
    elif pos >= len(proc_obj.stack):
        proc_obj.errmsg("Adjusting would put us beyond the newest frame.")
        return

    proc_obj.curindex = pos
    proc_obj.curframe = proc_obj.stack[proc_obj.curindex][0]
    proc_obj.location()
    proc_obj.list_lineno = None
    proc_obj.list_offset = proc_obj.curframe.f_lasti
    proc_obj.list_object = proc_obj.curframe
    proc_obj.list_filename = proc_obj.curframe.f_code.co_filename

    return


def adjust_relative(proc_obj, name, args, signum):
    if not proc_obj.stack:
        proc_obj.errmsg("Program has no stack frame set.")
        return False
    if len(args) == 1:
        count = 1
    else:
        count_str = args[1]
        low, high = frame_low_high(proc_obj, signum)
        count = Mcmdfns.get_an_int(
            proc_obj.errmsg,
            count_str,
            ("The '%s' command argument must eval to" + " an integer. Got: %s")
            % (name, count_str),
            low,
            high,
        )
        if count is None:
            return
        pass

    adjust_frame(proc_obj, name, pos=signum * count, absolute_pos=False)
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2013, 2015, 2017-2018, 2020-2021, 2023 Rocky
#   Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" Common command-parsing routines such as check command argument
counts, to parse a string for an integer, or check a string for an
on/off setting value.
"""
import sys
import tempfile

from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE


def source_tempfile_remap(prefix, text, tempdir=None):
    fd = tempfile.NamedTemporaryFile(
        suffix=".py", prefix=prefix, dir=tempdir, delete=False
    )
    with fd:
        fd.write(bytes(text, "UTF-8"))
        fd.close()
        pass
    return fd.name


def deparse_fn(code):
    try:
        if (3, 7) <= PYTHON_VERSION_TRIPLE <= (3, 8):
            from uncompyle6.semantics.linemap import (
                deparse_code_with_fragments_and_map as deparse_code,
            )
        else:
            from decompile3.semantics.linemap import (
                deparse_code_with_fragments_and_map as deparse_code,
            )

    except ImportError:
        return None
    try:
        deparsed = deparse_code(PYTHON_VERSION_TRIPLE, code, is_pypy=IS_PYPY)
        return deparsed
    except Exception:
        raise
    return None


def get_an_int(errmsg, arg, msg_on_error, min_value=None, max_value=None):
    """Another get_int() routine, this one simpler and less stylized
    than get_int(). We eval arg return it as an integer value or
    None if there was an error in parsing this.
    """
    ret_value = None
    if arg:
        try:
            # eval() is used so we will allow arithmetic expressions,
            # variables etc.
            ret_value = int(eval(arg))
        except (SyntaxError, NameError, ValueError):
            if errmsg:
                errmsg(msg_on_error)
            else:
                errmsg("Expecting an integer, got: %s." % str(arg))
            return None

    if min_value and ret_value < min_value:
        errmsg(
            "Expecting integer value to be at least %d, got: %d."
            % (min_value, ret_value)
        )
        return None
    elif max_value and ret_value > max_value:
        errmsg(
            "Expecting integer value to be at most %d, got: %d."
            % (max_value, ret_value)
        )
        return None
    return ret_value


def get_int(errmsg, arg, default=1, cmdname=None):
    """If arg is an int, use that otherwise take default."""
    if arg:
        try:
            # eval() is used so we will allow arithmetic expressions,
            # variables etc.
            default = int(eval(arg))
        except (SyntaxError, NameError, ValueError):
            if cmdname:
                errmsg(
                    "Command '%s' expects an integer; got: %s." % (cmdname, str(arg))
                )
            else:
                errmsg("Expecting an integer, got: %s." % str(arg))
                pass
            raise ValueError
    return default


def get_onoff(errmsg, arg, default=None, print_error=True):
    """Return True if arg is 'on' or 1 and False arg is 'off' or 0.
    Any other value is raises ValueError."""
    if not arg:
        if default is None:
            if print_error:
                errmsg("Expecting 'on', 1, 'off', or 0. Got nothing.")
                pass
            raise ValueError
        return default
    if arg == "1" or arg == "on":
        return True
    if arg == "0" or arg == "off":
        return False

    if print_error:
        errmsg("Expecting 'on', 1, 'off', or 0. Got: %s." % str(arg))
    raise ValueError


def get_val(curframe, errmsg, arg):
    try:
        return eval(arg, curframe.f_globals, curframe.f_locals)
    except Exception:
        t, v = sys.exc_info()[:2]
        if isinstance(t, str):
            exc_type_name = t
        else:
            exc_type_name = t.__name__
        errmsg(str("%s: %s" % (exc_type_name, arg)))
        raise
    return  # Not reached


def run_set_bool(obj, args):
    """set a Boolean-valued debugger setting. 'obj' is a generally a
    subcommand that has 'name' and 'debugger.settings' attributes"""
    try:
        if 0 == len(args):
            args = ["on"]
        obj.debugger.settings[obj.name] = get_onoff(obj.errmsg, args[0])
    except ValueError:
        pass
    return


def run_set_int(obj, arg, msg_on_error, min_value=None, max_value=None):
    """set an Integer-valued debugger setting. 'obj' is a generally a
    subcommand that has 'name' and 'debugger.settings' attributes"""
    if "" == arg.strip():
        obj.errmsg("You need to supply a number.")
        return
    obj.debugger.settings[obj.name] = get_an_int(
        obj.errmsg, arg, msg_on_error, min_value, max_value
    )
    return obj.debugger.settings[obj.name]


def run_show_bool(obj, what=None):
    """Generic subcommand showing a boolean-valued debugger setting.
    'obj' is generally a subcommand that has 'name' and
    'debugger.setting' attributes."""
    val = show_onoff(obj.debugger.settings[obj.name])
    if not what:
        what = obj.name
    return obj.msg("%s is %s." % (what, val))


def run_show_int(obj, what=None):
    """Generic subcommand integer value display"""
    val = obj.debugger.settings[obj.name]
    if not what:
        what = obj.name
    return obj.msg("%s is %d." % (what, val))


def show_onoff(b):
    """Return 'on' for True and 'off' for False, and ?? for anything
    else."""
    if not isinstance(b, bool):
        return "??"
    if b:
        return "on"
    return "off"


def run_show_val(obj, name):
    """Generic subcommand value display"""
    val = obj.debugger.settings[obj.name]
    obj.msg(
        "%s is %s."
        % (
            obj.name,
            obj.cmd.proc._saferepr(val),
        )
    )
    return False


def want_different_line(cmd, default):
    if cmd[-1] == "-":
        return False
    elif cmd[-1] == "+":
        return True
    return default


# Demo it
if __name__ == "__main__":

    def errmsg(msg):
        print("** %s" % msg)
        return

    def msg(m):
        print(m)

    print(get_int(errmsg, "1+2"))  # 3
    print(get_int(errmsg, None))  # 1
    print(get_an_int(errmsg, "6*1", "6*1 is okay"))  # 6
    print(get_an_int(errmsg, "0", "0 is too small", 1))  # errmsg
    print(get_an_int(errmsg, "5+a", "5+a is no good"))  # errmsg
    try:
        get_int(errmsg, "pi")
    except ValueError:
        print("Good - 'pi' is not an integer")
        pass

    import inspect

    curframe = inspect.currentframe()

    print(want_different_line("s+ %s", False))
    print(want_different_line("s-", True))
    print(want_different_line("s", False))
    print(want_different_line("s", True))
    print(want_different_line("s", True))
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2014, 2016-2017 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger Server Input/Output interface. """

import socket, errno

from trepan.lib import default as Mdefault
from trepan import misc as Mmisc
from trepan.inout import tcpfns as Mtcpfns
from trepan.inout.base import DebuggerInOutBase


# FIXME: Consider using Python's socketserver/SocketServer?
class TCPServer(DebuggerInOutBase):
    """Debugger Server Input/Output Socket."""

    DEFAULT_INIT_OPTS = {"open": True, "socket": None}

    def __init__(self, inout=None, opts=None):
        get_option = lambda key: Mmisc.option_set(opts, key, self.DEFAULT_INIT_OPTS)

        self.inout = None
        self.conn = None
        self.addr = None
        self.remote_addr = ""
        self.buf = ""  # Read buffer
        self.line_edit = False  # Our name for GNU readline capability
        self.state = "disconnected"
        self.PORT = None
        self.HOST = None
        if inout:
            self.inout = inout
        if get_option("socket"):
            self.inout = opts["socket"]
            self.inout.listen(1)
            self.state = "listening"
        elif get_option("open"):
            self.open(opts)
            pass
        return

    def close(self):
        """Closes both socket and server connection."""
        self.state = "closing"
        if self.inout:
            self.inout.close()
            pass
        self.state = "closing connection"
        if self.conn:
            self.conn.close()
        self.state = "disconnected"
        return

    def open(self, opts=None):
        get_option = lambda key: Mmisc.option_set(
            opts, key, Mdefault.SERVER_SOCKET_OPTS
        )

        self.HOST = get_option("HOST")
        self.PORT = get_option("PORT")
        self.reuse = get_option("reuse")
        self.search_limit = get_option("search_limit")
        self.inout = None

        this_port = self.PORT - 1
        for i in range(self.search_limit):
            this_port += 1
            for res in socket.getaddrinfo(
                self.HOST,
                this_port,
                socket.AF_UNSPEC,
                socket.SOCK_STREAM,
                0,
                socket.AI_PASSIVE,
            ):
                af, socktype, proto, canonname, sa = res
                try:
                    self.inout = socket.socket(af, socktype, proto)
                except socket.error:
                    self.inout = None
                    continue
                try:
                    if get_option("reuse"):
                        # The following socket option allows the OS to reclaim
                        # The address faster on termination.
                        self.inout.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

                        pass
                    self.inout.bind(sa)
                    self.inout.listen(1)
                    self.state = "listening"
                    break
                except socket.error as exc:
                    if exc.errno in [errno.EADDRINUSE, errno.EINVAL]:
                        self.inout.close()
                        self.inout = None
                        continue
                    raise
                pass
            if self.state == "listening":
                break
        if self.inout is None:
            raise IOError(
                "could not open server socket after trying ports "
                "%s..%s" % (self.PORT, this_port)
            )
        self.PORT = this_port
        return

    def read(self):
        if len(self.buf) == 0:
            self.read_msg()
            pass
        return self.buf

    def read_msg(self):
        """Read one message unit. It's possible however that
        more than one message will be set in a receive, so we will
        have to buffer that for the next read.
        EOFError will be raised on EOF.
        """
        if self.state != "connected":
            self.wait_for_connect()
            pass
        if self.state == "connected":
            if 0 == len(self.buf):
                self.buf = self.conn.recv(Mtcpfns.TCP_MAX_PACKET)
                if 0 == len(self.buf):
                    self.state = "disconnected"
                    raise EOFError
                pass
            self.buf, data = Mtcpfns.unpack_msg(self.buf)
            return data.decode("utf-8")
        else:
            raise IOError("read_msg called in state: %s." % self.state)

    def wait_for_connect(self):
        self.conn, self.addr = self.inout.accept()
        self.remote_addr = ":".join(str(v) for v in self.addr)
        self.state = "connected"
        return

    def write(self, msg):
        """This method the debugger uses to write. In contrast to
        writeline, no newline is added to the end to `str'. Also
        msg doesn't have to be a string.
        """
        if self.state != "connected":
            self.wait_for_connect()
            pass
        buffer = Mtcpfns.pack_msg(msg)
        while len(buffer) > Mtcpfns.TCP_MAX_PACKET:
            self.conn.send(buffer[: Mtcpfns.TCP_MAX_PACKET])
            buffer = buffer[Mtcpfns.TCP_MAX_PACKET :]
        return self.conn.send(buffer)


# Demo
if __name__ == "__main__":
    inout = TCPServer(opts={"open": False})
    import sys

    if len(sys.argv) > 1:
        inout.open()
        print("Listening for connection on %s:%s" % (inout.HOST, inout.PORT))
        while True:
            try:
                line = inout.read_msg().rstrip("\n")
                print(line)
                inout.writeline("ack: " + line)
            except EOFError:
                break
            pass
        pass
    inout.close()
    pass



#!/usr/bin/env python
"General integration tests"
import unittest

import helper as Mhelper


class NoScriptTests(unittest.TestCase):

    def test_basic(self):
        """Test invoking debugger without a file name"""
        result=Mhelper.run_debugger(testname='noscript',
                                    dbgr_opts='--basename',
                                    python_file=None)
        self.assertEqual(True, result, "debugger 'step' command comparision")
        return
    pass

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009-2010, 2013, 2015, 2017, 2020, 2023 Rocky
#  Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import code
import sys

from traitlets.config.loader import Config

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class IPythonCommand(DebuggerCommand):
    """**ipython** [**-d**]

    Run IPython as a command subshell.

    If *-d* is passed, you can access debugger state via local variable *debugger*.

    To issue a debugger command use function *dbgr()*. For example:

      dbgr('info program')

    See also:
    ---------

    `python`, `bpython`"""

    short_help = "Run IPython as a command subshell"

    DebuggerCommand.setup(locals(), category="support", max_args=1)

    def dbgr(self, string):
        """Invoke a debugger command from inside a IPython shell called
        inside the debugger.
        """
        self.proc.cmd_queue.append(string)
        self.proc.process_command()
        return

    def run(self, args):
        # See if python's code module is around

        # Python does it's own history thing.
        # Make sure it doesn't damage ours.
        have_line_edit = self.debugger.intf[-1].input.line_edit
        if have_line_edit:
            try:
                self.proc.write_history_file()
            except IOError:
                pass
            pass

        cfg = Config()
        banner_tmpl = """IPython trepan3k shell%s

Use dbgr(*string*) to issue non-continuing debugger command: *string*"""

        debug = len(args) > 1 and args[1] == "-d"
        if debug:
            banner_tmpl += "\nVariable 'debugger' contains a trepan " "debugger object."
            pass
        try:
            from IPython.terminal.embed import InteractiveShellEmbed
        except ImportError:
            from IPython.frontend.terminal.embed import InteractiveShellEmbed

        # Now create an instance of the embeddable shell. The first
        # argument is a string with options exactly as you would type them
        # if you were starting IPython at the system command line. Any
        # parameters you want to define for configuration can thus be
        # specified here.

        # Add common classes and methods our namespace here so that
        # inside the ipython shell users don't have run imports

        my_locals = {}
        my_globals = None
        if self.proc.curframe:
            my_globals = self.proc.curframe.f_globals
            if self.proc.curframe.f_locals:
                my_locals = self.proc.curframe.f_locals
            pass
        pass

        # Give IPython and the user a way to get access to the debugger.
        if debug:
            my_locals["debugger"] = self.debugger
        my_locals["dbgr"] = self.dbgr
        cfg.TerminalInteractiveShell.confirm_exit = False

        # sys.ps1 = 'trepan3 >>> '
        if len(my_locals):
            banner = banner_tmpl % " with locals"
        else:
            banner = banner_tmpl % ""
            pass

        InteractiveShellEmbed(
            config=cfg,
            banner1=banner,
            user_ns=my_locals,
            module=my_globals,
            exit_msg="IPython exiting to trepan3k...",
        )()

        # restore completion and our history if we can do so.
        if hasattr(self.proc.intf[-1], "complete"):
            try:
                from readline import parse_and_bind, set_completer

                parse_and_bind("tab: complete")
                set_completer(self.proc.intf[-1].complete)
            except ImportError:
                pass
            pass

        if have_line_edit:
            self.proc.read_history_file()
            pass
        return

    pass


# Monkey-patched from code.py
# FIXME: get changes into Python.
def interact(banner=None, readfunc=None, my_locals=None, my_globals=None):
    """Almost a copy of code.interact
    Closely emulate the interactive Python interpreter.

    This is a backwards compatible interface to the InteractiveConsole
    class.  When readfunc is not specified, it attempts to import the
    readline module to enable GNU readline if it is available.

    Arguments (all optional, all default to None):

    banner -- passed to InteractiveConsole.interact()
    readfunc -- if not None, replaces InteractiveConsole.raw_input()
    local -- passed to InteractiveInterpreter.__init__()

    """
    console = code.InteractiveConsole(my_locals, filename="<trepan>")
    console.runcode = lambda code_obj: runcode(console, code_obj)
    setattr(console, "globals", my_globals)
    if readfunc is not None:
        console.raw_input = readfunc
    console.interact(banner)
    pass


# Also monkey-patched from code.py
# FIXME: get changes into Python.
def runcode(obj, code_obj):
    """Execute a code object.

    When an exception occurs, self.showtraceback() is called to
    display a traceback.  All exceptions are caught except
    SystemExit, which is reraised.

    A note about KeyboardInterrupt: this exception may occur
    elsewhere in this code, and may not always be caught.  The
    caller should be prepared to deal with it.

    """
    try:
        exec(code_obj, obj.locals, obj.globals)
    except SystemExit:
        raise
    except Exception:
        obj.showtraceback()
    else:
        if code.softspace(sys.stdout, 0):
            print()
            pass
        pass
    return


if __name__ == "__main__":
    from trepan.debugger import Trepan

    d = Trepan()
    command = IPythonCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()
    if len(sys.argv) > 1:
        print("Type Python commands and exit to quit.")
        print(sys.argv[1])
        if sys.argv[1] == "-d":
            print(command.run(["bpython", "-d"]))
        else:
            print(command.run(["bpython"]))
            pass
        pass
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2013-2014, 2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Simulate I/O using lists of strings. """

import types

from trepan.inout import base as Mbase


class StringArrayInput(Mbase.DebuggerInputBase):
    """Simulate I/O using an array of strings. Sort of like StringIO, but
    even simpler."""

    def __init__(self, inp=[], opts=None):
        self.input = inp
        self.closed = False
        return

    def close(self):
        "Interface is for compatibility"
        self.closed = True
        return

    def open(self, inp, opts=None):
        """Use this to set where to read from."""
        if isinstance(inp, list):
            self.input = inp
        else:
            raise IOError("Invalid input type (%s) for %s" % (type(inp), inp))
        return

    def readline(self, use_raw=None, prompt=""):
        """Read a line of input. EOFError will be raised on EOF.

        Note that we don't support prompting"""
        if self.closed:
            raise ValueError
        if 0 == len(self.input):
            self.closed = True
            raise EOFError
        line = self.input[0]
        del self.input[0]
        return line

    pass


class StringArrayOutput(Mbase.DebuggerInOutBase):
    """Simulate I/O using an array of strings. Sort of like StringIO, but
    even simpler."""

    def __init__(self, out=[], opts=None):
        self.flush_after_write = False  # For compatibility
        self.closed = False
        self.output = out
        return

    def close(self):
        "Nothing to do here. Interface is for compatibility"
        self.closed = True
        return

    def flush(self):
        "Nothing to do here. Interface is for compatibility"
        return

    def open(self, output):
        """Use this to set where to write to. output can be a
        file object or a string. This code raises IOError on error.

        If another file was previously open upon calling this open,
        that will be stacked and will come back into use after
        a close_write().
        """
        if isinstance(output, types.Listype):
            self.output = output
        else:
            raise IOError("Invalid output type (%s) for %s" % (type(output), output))
        return

    def write(self, msg):
        """This method the debugger uses to write. In contrast to
        writeline, no newline is added to the end to `str'.
        """
        if self.closed:
            raise ValueError
        if [] == self.output:
            self.output = [msg]
        else:
            self.output[-1] += msg
            pass
        return

    def writeline(self, msg):
        """used to write to a debugger that is connected to this
        server; Here, we use the null string '' as an indicator of a
        newline.
        """
        self.write(msg)
        self.output.append("")
        return

    pass


# Demo
if __name__ == "__main__":
    inp = StringArrayInput(["Now is the time", "for all good men"])
    line = inp.readline()
    print(line)
    line = inp.readline()
    print(line)
    try:
        line = inp.readline()
    except EOFError:
        print("EOF hit on read")
        pass
    out = StringArrayOutput()
    print(out.output)
    #    line = io.readline("Type some more characters: ")
    out.writeline("Hello, world!")
    print(out.output)
    out.write("Hello")
    print(out.output)
    out.writeline(", again.")
    print(out.output)
    #     io.open_write(sys.stdout)
    out.flush_after_write = True
    out.write("Last hello")
    out.close()
    print(out.output)
    try:
        out.writeline("You won't see me")
    except Exception:
        pass
    # Closing after already closed is okay
    out.close()
    inp.close()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2014, 2016, 2022, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Signal handlers."""
# TODO:
#  - Doublecheck handle_pass and other routines.
#  - can remove signal handler altogether when
#         ignore=True, print=False, pass=True
#
#
import signal
from typing import Optional


def yes_or_no(b) -> str:
    """Return 'Yes' for True and 'No' for False, and ?? for anything
    else."""
    if type(b) != bool:
        return "??"
    if b:
        return "Yes"
    return "No"


def lookup_signame(num: int) -> Optional[str]:
    """Find the corresponding signal name for 'num'. Return None
    if 'num' is invalid."""
    signames = signal.__dict__
    num = abs(num)
    for signame in list(signames.keys()):
        if signame.startswith("SIG") and signames[signame] == num:
            return signame
        pass
    # Something went wrong. Should have returned above
    return None


def lookup_signum(name):
    """Find the corresponding signal number for 'name'. Return None
    if 'name' is invalid."""
    uname = name.upper()
    if uname.startswith("SIG") and hasattr(signal, uname):
        return getattr(signal, uname)
    else:
        uname = "SIG" + uname
        if hasattr(signal, uname):
            return getattr(signal, uname)
        return None
    return  # noqa


def canonic_signame(name_num) -> Optional[str]:
    """Return a signal name for a signal name or signal
    number.  Return None is name_num is an int but not a valid signal
    number and False if name_num is a not number. If name_num is a
    signal name or signal number, the canonic if name is returned."""
    signum = lookup_signum(name_num)
    if signum is None:
        # Maybe signame is a number?
        try:
            num = int(name_num)
            signame = lookup_signame(num)
            if signame is None:
                return None
        except Exception:
            return False
        return signame

    signame = name_num.upper()
    if not signame.startswith("SIG"):
        return "SIG" + signame
    return signame


fatal_signals = ["SIGKILL", "SIGSTOP"]

# I copied these from GDB source code.
signal_description = {
    "SIGHUP": "Hangup",
    "SIGINT": "Interrupt",
    "SIGQUIT": "Quit",
    "SIGILL": "Illegal instruction",
    "SIGTRAP": "Trace/breakpoint trap",
    "SIGABRT": "Aborted",
    "SIGEMT": "Emulation trap",
    "SIGFPE": "Arithmetic exception",
    "SIGKILL": "Killed",
    "SIGBUS": "Bus error",
    "SIGSEGV": "Segmentation fault",
    "SIGSYS": "Bad system call",
    "SIGPIPE": "Broken pipe",
    "SIGALRM": "Alarm clock",
    "SIGTERM": "Terminated",
    "SIGURG": "Urgent I/O condition",
    "SIGSTOP": "Stopped (signal)",
    "SIGTSTP": "Stopped (user)",
    "SIGCONT": "Continued",
    "SIGCHLD": "Child status changed",
    "SIGTTIN": "Stopped (tty input)",
    "SIGTTOU": "Stopped (tty output)",
    "SIGIO": "I/O possible",
    "SIGXCPU": "CPU time limit exceeded",
    "SIGXFSZ": "File size limit exceeded",
    "SIGVTALRM": "Virtual timer expired",
    "SIGPROF": "Profiling timer expired",
    "SIGWINCH": "Window size changed",
    "SIGLOST": "Resource lost",
    "SIGUSR1": "User-defined signal 1",
    "SIGUSR2": "User-defined signal 2",
    "SIGPWR": "Power fail/restart",
    "SIGPOLL": "Pollable event occurred",
    "SIGWIND": "SIGWIND",
    "SIGPHONE": "SIGPHONE",
    "SIGWAITING": "Process's LWPs are blocked",
    "SIGLWP": "Signal LWP",
    "SIGDANGER": "Swap space dangerously low",
    "SIGGRANT": "Monitor mode granted",
    "SIGRETRACT": "Need to relinquish monitor mode",
    "SIGMSG": "Monitor mode data available",
    "SIGSOUND": "Sound completed",
    "SIGSAK": "Secure attention",
}


class SignalManager:
    """Manages Signal Handling information for the debugger

    - Do we print/not print when signal is caught
    - Do we pass/not pass the signal to the program
    - Do we stop/not stop when signal is caught

    Parameter dbgr is a Debugger object. ignore is a list of
    signals to ignore. If you want no signals, use [] as None uses the
    default set. Parameter default_print specifies whether or not we
    print receiving a signals that is not ignored.

    All the methods which change these attributes return None on error, or
    True/False if we have set the action (pass/print/stop) for a signal
    handler.
    """

    def __init__(self, dbgr, ignore_list=None, default_print=True):
        self.dbgr = dbgr
        # dbgr.core.add_ignore(SigHandler.handle)
        self.sigs = {}

        # List of signals. Dunno why signal doesn't provide.
        self.siglist = []

        # Ignore signal handling initially for these known signals.
        if ignore_list is None:
            ignore_list = [
                "SIGALRM",
                "SIGCHLD",
                "SIGURG",
                "SIGIO",
                "SIGCLD",
                "SIGVTALRM" "SIGPROF",
                "SIGWINCH",
                "SIGPOLL",
                "SIGWAITING",
                "SIGLWP",
                "SIGCANCEL",
                "SIGTRAP",
                "SIGTERM",
                "SIGQUIT",
                "SIGILL",
                # Weird stuff from 3.3
                "SIG_SETMASK",
                "ITIMER_PROF",
                "ITIMER_VIRTUAL",
                "ITIMER_REAL",
                "ITIMER_PROF",
                "SIG_BLOCK",
                "SIG_UNBLOCK",
            ]  # NOQA
        self.ignore_list = ignore_list
        self._orig_set_signal = signal.signal
        signal.signal = self.set_signal_replacement

        self.info_fmt = "%-14s%-4s\t%-4s\t%-5s\t%-4s\t%s"
        self.header = self.info_fmt % (
            "Signal",
            "Stop",
            "Print",
            "Stack",
            "Pass",
            "Description",
        )

        if default_print:
            default_print = self.dbgr.intf[-1].msg

        for signame in list(signal.__dict__.keys()):
            # Look for a signal name on this os.
            if signame.startswith("SIG") and "_" not in signame:
                self.siglist.append(signame)
                self.initialize_handler(signame)
            pass
        self.action("SIGINT stop print nostack nopass")
        return

    def initialize_handler(self, signame: str):
        if signame in fatal_signals:
            return False
        signum = lookup_signum(signame)
        if signum is None:
            return False

        try:
            old_handler = signal.getsignal(signum)
        except ValueError:
            # On some OS's (Redhat 8), SIGNUM's are listed (like
            # SIGRTMAX) that getsignal can't handle.
            old_handler = None
            if signame in self.sigs:
                self.sigs.pop(signame)
                pass

        if signame in self.ignore_list:
            self.sigs[signame] = SigHandler(
                self.dbgr,
                signame,
                signum,
                old_handler,
                None,
                False,
                print_stack=False,
                pass_along=True,
            )
        else:
            self.sigs[signame] = SigHandler(
                self.dbgr,
                signame,
                signum,
                old_handler,
                self.dbgr.intf[-1].msg,
                True,
                print_stack=False,
                pass_along=False,
            )
            pass
        return True

    def set_signal_replacement(self, signum: int, handle):
        """A replacement for signal.signal which chains the signal behind
        the debugger's handler"""
        signame = lookup_signame(signum)
        if signame is None:
            self.dbgr.intf[-1].errmsg(
                "%s is not a signal number I know about." % signum
            )
            return False
        # Since the intent is to set a handler, we should pass this
        # signal on to the handler
        if signame not in self.ignore_list:
            self.sigs[signame].pass_along = True
            if self.check_and_adjust_sighandler(signame, self.sigs):
                self.sigs[signame].old_handler = handle
                return True
        return False

    def check_and_adjust_sighandler(self, signame: str, sigs):
        """
        Check to see if a single signal handler that we are interested
        in has changed or has not been set initially. On return
        self.sigs[signame] should have our signal handler. True is
        returned if the same or adjusted, False or None if error or
        not found.
        """

        signum = lookup_signum(signame)

        try:
            old_handler = signal.getsignal(signum)
        except ValueError:
            # On some OS's (Redhat 8), SIGNUM's are listed (like
            # SIGRTMAX) that getsignal can't handle.
            if signame in self.sigs:
                sigs.pop(signame)
                pass
            return None
        if old_handler != self.sigs[signame].handle:
            if old_handler not in [signal.SIG_IGN, signal.SIG_DFL]:
                # save the program's signal handler
                sigs[signame].old_handler = old_handler
                pass
            # set/restore _our_ signal handler
            try:
                # signal.signal(signum, self.sigs[signame].handle)
                self._orig_set_signal(signum, self.sigs[signame].handle)
            except ValueError:
                # Probably not in main thread
                return False
            except KeyError:
                # May be weird keys from 3.3
                return False
            pass
        return True

    def check_and_adjust_sighandlers(self):
        """Check to see if any of the signal handlers we are interested in have
        changed or is not initially set. Change any that are not right."""
        for signame in list(self.sigs.keys()):
            if not self.check_and_adjust_sighandler(signame, self.sigs):
                break
            pass
        return

    def is_name_or_number(self, name_num):
        signame = canonic_signame(name_num)
        if signame is None:
            self.dbgr.intf[-1].errmsg(
                ("%s is not a signal number" + " I know about.") % name_num
            )
            return False
        elif not signame:
            self.dbgr.intf[-1].errmsg(
                ("%s is not a signal name I " + "know about.") % name_num
            )
            return False
        return signame

    def print_info_signal_entry(self, signame):
        """Print status for a single signal name (signame)"""
        if signame in signal_description:
            description = signal_description[signame]
        else:
            description = ""
            pass
        if signame not in list(self.sigs.keys()):
            # Fake up an entry as though signame were in sigs.
            self.dbgr.intf[-1].msg(
                self.info_fmt % (signame, "No", "No", "No", "Yes", description)
            )
            return

        sig_obj = self.sigs[signame]
        self.dbgr.intf[-1].msg(
            self.info_fmt
            % (
                signame,
                yes_or_no(sig_obj.b_stop),
                yes_or_no(sig_obj.print_method is not None),
                yes_or_no(sig_obj.print_stack),
                yes_or_no(sig_obj.pass_along),
                description,
            )
        )
        return

    def info_signal(self, args):
        """Print information about a signal"""
        if len(args) == 0:
            return None
        signame = args[0]
        if signame in ["handle", "signal"]:
            # This has come from dbgr's info command
            if len(args) == 1:
                # Show all signal handlers
                self.dbgr.core.processor.section(self.header)
                for signame in self.siglist:
                    self.print_info_signal_entry(signame)
                return True
            else:
                signame = args[1]
                pass
            pass

        signame = self.is_name_or_number(signame)
        self.dbgr.core.processor.section(self.header)
        self.print_info_signal_entry(signame)
        return True

    def action(self, arg):
        """Delegate the actions specified in 'arg' to another
        method.
        """
        if not arg:
            self.info_signal(["handle"])
            return True
        args = arg.split()
        signame = args[0]
        signame = self.is_name_or_number(args[0])
        if not signame:
            return

        if len(args) == 1:
            self.info_signal([signame])
            return True
        # We can display information about 'fatal' signals, but not
        # change their actions.
        if signame in fatal_signals:
            return None

        if signame not in list(self.sigs.keys()):
            if not self.initialize_handler(signame):
                return None
            pass

        # multiple commands might be specified, i.e. 'nopass nostop'
        for attr in args[1:]:
            if attr.startswith("no"):
                on = False
                attr = attr[2:]
            else:
                on = True
            if "stop".startswith(attr):
                self.handle_stop(signame, on)
            elif "print".startswith(attr) and len(attr) >= 2:
                self.handle_print(signame, on)
            elif "pass".startswith(attr):
                self.handle_pass(signame, on)
            elif "ignore".startswith(attr):
                self.handle_ignore(signame, on)
            elif "stack".startswith(attr):
                self.handle_print_stack(signame, on)
            else:
                self.dbgr.intf[-1].errmsg("Invalid arguments")
                pass
            pass
        return self.check_and_adjust_sighandler(signame, self.sigs)

    def handle_print_stack(self, signame, print_stack):
        """Set whether we stop or not when this signal is caught.
        If 'set_stop' is True your program will stop when this signal
        happens."""
        self.sigs[signame].print_stack = print_stack
        return print_stack

    def handle_stop(self, signame, set_stop):
        """Set whether we stop or not when this signal is caught.
        If 'set_stop' is True your program will stop when this signal
        happens."""
        if set_stop:
            self.sigs[signame].b_stop = True
            # stop keyword implies print AND nopass
            self.sigs[signame].print_method = self.dbgr.intf[-1].msg
            self.sigs[signame].pass_along = False
        else:
            self.sigs[signame].b_stop = False
            pass
        return set_stop

    def handle_pass(self, signame, set_pass):
        """Set whether we pass this signal to the program (or not)
        when this signal is caught. If set_pass is True, Dbgr should allow
        your program to see this signal.
        """
        self.sigs[signame].pass_along = set_pass
        if set_pass:
            # Pass implies nostop
            self.sigs[signame].b_stop = False
            pass
        return set_pass

    def handle_ignore(self, signame, set_ignore):
        """'pass' and 'noignore' are synonyms. 'nopass and 'ignore' are
        synonyms."""
        self.handle_pass(signame, not set_ignore)
        return set_ignore

    def handle_print(self, signame, set_print):
        """Set whether we print or not when this signal is caught."""
        if set_print:
            self.sigs[signame].print_method = self.dbgr.intf[-1].msg
        else:
            self.sigs[signame].print_method = None
            pass
        return set_print

    pass


class SigHandler:
    """Store information about what we do when we handle a signal,

    - Do we print/not print when signal is caught
    - Do we pass/not pass the signal to the program
    - Do we stop/not stop when signal is caught

    Parameters:
       signame : name of signal (e.g. SIGUSR1 or USR1)
       print_method routine to use for "print"
       stop routine to call to invoke debugger when stopping
       pass_along: True is signal is to be passed to user's handler
    """

    def __init__(
        self,
        dbgr,
        signame,
        signum,
        old_handler,
        print_method,
        b_stop,
        print_stack=False,
        pass_along=True,
    ):

        self.dbgr = dbgr
        self.old_handler = old_handler
        self.pass_along = pass_along
        self.print_method = print_method
        self.print_stack = print_stack
        self.signame = signame
        self.signum = signum
        self.b_stop = b_stop
        return

    def handle(self, signum, frame):
        """This method is called when a signal is received."""
        if self.print_method:
            self.print_method("\nProgram received signal %s." % self.signame)
        if self.print_stack:
            import traceback

            strings = traceback.format_stack(frame)
            for s in strings:
                if s[-1] == "\n":
                    s = s[0:-1]
                self.print_method(s)
                pass
            pass
        if self.b_stop:
            core = self.dbgr.core
            old_trace_hook_suspend = core.trace_hook_suspend
            core.trace_hook_suspend = True
            core.stop_reason = "intercepting signal %s (%d)" % (self.signame, signum)
            core.processor.event_processor(frame, "signal", signum)
            core.trace_hook_suspend = old_trace_hook_suspend
            pass
        if self.pass_along:
            # pass the signal to the program
            if self.old_handler:
                self.old_handler(signum, frame)
                pass
            pass
        return

    pass


# When invoked as main program, do some basic tests of a couple of functions
if __name__ == "__main__":

    for b in (
        True,
        False,
    ):
        print("yes_or_no of %s is %s" % (repr(b), yes_or_no(b)))
        pass
    for signum in range(signal.NSIG):
        signame = lookup_signame(signum)
        if signame is not None:
            if not signame.startswith("SIG"):
                continue
            print(signame, signum, lookup_signum(signame))
            assert signum == lookup_signum(signame)
            # Try without the SIG prefix
            assert signum == lookup_signum(signame[3:])
            pass
        pass

    for i in (15, -15, 300):
        print("lookup_signame(%d): %s" % (i, lookup_signame(i)))
        pass

    for i in ("term", "TERM", "NotThere"):
        print("lookup_signum(%s): %s" % (i, repr(lookup_signum(i))))
        pass

    for i in ("15", "-15", "term", "sigterm", "TERM", "300", "bogus"):
        print("canonic_signame(%s): %s" % (i, canonic_signame(i)))
        pass

    from trepan import debugger as Mdebugger

    dbgr = Mdebugger.Trepan()
    h = SignalManager(dbgr)
    h.info_signal(["TRAP"])
    # Set to known value
    h.action("SIGUSR1")
    h.action("usr1 print pass stop")
    h.info_signal(["USR1"])
    # noprint implies no stop
    h.action("SIGUSR1 noprint")
    h.info_signal(["USR1"])
    h.action("foo nostop")
    # stop keyword implies print
    h.action("SIGUSR1 stop")
    h.info_signal(["SIGUSR1"])
    h.action("SIGUSR1 noprint")
    h.info_signal(["SIGUSR1"])
    h.action("SIGUSR1 nopass stack")
    h.info_signal(["SIGUSR1"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.frame import adjust_relative, frame_complete


class UpCommand(DebuggerCommand):

    signum = -1  # This is what distinguishes us from "down"

    DebuggerCommand.setup(locals(), category="stack", need_stack=True, max_args=1)

    short_help = (
        "Move frame in the direction of the caller of " "the last-selected frame"
    )

    def complete(self, prefix):
        proc_obj = self.proc
        return frame_complete(proc_obj, prefix, self.signum)

    def run(self, args):
        """**up** [*count*]

        Move the current frame up in the stack trace (to an older frame). 0 is
        the most recent frame. If no count is given, move up 1.

        See also:
        ---------

        `down` and `frame`."""
        adjust_relative(self.proc, self.name, args, self.signum)
        return False


if __name__ == "__main__":
    from trepan.processor import cmdproc as Mcmdproc
    from trepan import debugger as Mdebugger

    d = Mdebugger.Trepan()
    cp = d.core.processor
    command = UpCommand(cp)
    command.run(["up"])

    def nest_me(cp, command, i):
        import inspect

        if i > 1:
            cp.curframe = inspect.currentframe()
            cp.stack, cp.curindex = Mcmdproc.get_stack(cp.curframe, None, None, cp)
            command.run(["up"])
            print("-" * 10)
            command.run(["up", "-2"])
            print("-" * 10)
            command.run(["up", "-3"])
            print("-" * 10)
        else:
            nest_me(cp, command, i + 1)
        return

    cp.forget()
    nest_me(cp, command, 1)
    pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009-2010, 2013-2015, 2017-2018, 2020, 2023 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.processor.cmdbreak import parse_break_cmd, set_break

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete import complete_break_linenumber


class BreakCommand(DebuggerCommand):
    """**break** [*location*] [if *condition*]]

    Sets a breakpoint, i.e. stopping point just before the
    execution of the instruction specified by *location*.

    Without arguments or an empty *location*, the breakpoint is set at the
    current stopped location.

    See `help syntax location` for detailed information on a location.

    If the word `if` is given after *location*, subsequent arguments given
    Without arguments or an empty *location*, the breakpoint is set
    the current stopped location.

    Normally we only allow stopping at lines that we think are
    stoppable. If the command has a `!` suffix, force the breakpoint anyway.

    Examples:
    ---------

       break                # Break where we are current stopped at
       break if i < j       # Break at current line if i < j
       break 10             # Break on line 10 of the file we are
                            # currently stopped at
       break! 10            # Break where we are current stopped at, even if
                            # we don't think line 10 is stoppable
       break os.path.join() # Break in function os.path.join
       break x[i].fn()      # break in function specified by x[i].fn
       break x[i].fn() if x # break in function specified by x[i].fn
                            # if x is set
       break os.path:45     # Break on line 45 file holding module os.path
       break myfile.py:2    # Break on line 2 of myfile.py
       break myfile.py:2 if i < j # Same as above but only if i < j
       break "foo's.py":1"  # One way to specify path with a quote
       break 'c:\\foo.bat':1      # One way to specify a Windows file name,
       break '/My Docs/foo.py':1  # One way to specify path with blanks in it

    See also:
    ---------

    `info break`, `tbreak`, `condition` and `help syntax location`."""

    aliases = ("b", "break!", "b!")
    short_help = "Set breakpoint at specified line or function"

    DebuggerCommand.setup(locals(), category="breakpoints", need_stack=True)

    complete = complete_break_linenumber

    def run(self, args):
        force = True if args[0][-1] == "!" else False

        (func, filename, lineno, condition, offset) = parse_break_cmd(self.proc, args)
        if not (func is None and filename is None):
            set_break(
                self,
                func,
                filename,
                lineno,
                condition,
                False,
                args,
                force=force,
                offset=offset,
            )
        return


if __name__ == "__main__":

    def do_parse(cmd, a):
        name = "break"
        cmd.proc.current_command = name + " " + " ".join(a)
        print(a, "\n\t", parse_break_cmd(cmd.proc, [name] + a))

    def do_run(cmd, a):
        name = "break"
        cmd.proc.current_command = name + " " + " ".join(a)
        print(a)
        cmd.run([name] + a)

    import sys

    from trepan.debugger import Trepan

    d = Trepan()
    command = BreakCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()

    # do_parse(command, [""])
    # import inspect
    # line = inspect.currentframe().f_lineno
    # do_parse(command, [str(line)])
    # do_parse(command, ["%s:%s" % (__file__, line)])

    # def foo():
    #     return "bar"

    # do_parse(command, ["foo()"])
    # do_parse(command, ["os.path"])
    # do_parse(command, ["os.path", "5"])
    # import os.path
    # do_parse(command, ["os.path.join()"])
    # do_parse(command, ["if", "True"])
    # do_parse(command, ["foo()", "if", "True"])
    # do_parse(command, ["os.path:10", "if", "True"])

    do_run(command, [""])
    # do_run(command, ["command.run()"])
    do_run(command, ["89"])
    # command.run(["break", __file__ + ":10"])
    # command.run(["break", "foo"])
    pass



#!/usr/bin/env python3
'Unit test for trepan.lib.thread'
import sys, _thread, threading, unittest

from trepan.lib import thred as Mthread

class BgThread(threading.Thread):
    def __init__(self, id_name_checker):
        threading.Thread.__init__(self)
        self.id_name_checker = id_name_checker
        return

    def run(self):
        self.id_name_checker()
        return
    pass


class TestLibThread(unittest.TestCase):

    def id_name_checker(self):
        '''Helper for testing map_thread_names and id2thread'''
        name2id = Mthread.map_thread_names()
        for thread_id, f in list(sys._current_frames().items()):
            self.assertEqual(thread_id,
                             name2id[Mthread.id2thread_name(thread_id)])
            # FIXME: use a better test
            self.assertNotEqual(f, Mthread.find_debugged_frame(f))
            pass

    def test_current_thread_name(self):
        self.assertEqual('MainThread', Mthread.current_thread_name())
        return

    def test_id2thread_name(self):
        '''Test map_thread_names and id2thread'''
        thread_id = _thread.get_ident()
        self.assertEqual('MainThread', Mthread.id2thread_name(thread_id))
        self.id_name_checker()

        background = BgThread(self.id_name_checker)
        background.start()
        background.join()    # Wait for the background task to finish
        return
    pass

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2012-2013, 2016, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSetBoolSubcommand
from trepan.processor.cmdfns import run_set_bool, run_show_bool


class SetAutoPC(DebuggerSetBoolSubcommand):
    """**set autopc** [ **on** | **off** ]

    Run the `info pc` command every time we enter the debugger.

    See also:
    ---------

    `show autopc`"""

    in_list = True
    min_abbrev = len("autop")

    infopc_cmd = None

    def run(self, args):
        run_set_bool(self, args)
        if self.settings["autopc"]:
            if self.infopc_cmd is None:
                info_cmd = self.proc.commands["info"]
                self.info_pc_cmd = info_cmd.cmds.lookup("pc").run
                pass
            self.proc.add_preloop_hook(self.run_infopc, 0)

        else:
            self.proc.remove_preloop_hook(self.run_infopc)
            pass
        run_show_bool(self, "Run `info pc` on debugger entry")
        return

    def run_infopc(self, args):
        if self.proc.frame:
            self.info_pc_cmd(["info", "pc"])
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.show_subcmd.__demo_helper__ import demo_run

    demo_run(SetAutoPC)
    pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009-2010, 2012-2013, 2015, 2021, 2023 Rocky
#  Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""A base class for debugger commands.

This file is the one module in this directory that isn't a real command
and commands.py needs to take care to avoid instantiating this class
and storing it as a list of known debugger commands.
"""

import columnize
from pygments.console import colorize

from trepan.lib import format as Mformat

NotImplementedMessage = "This method must be overridden in a subclass"

__all__ = ["DebuggerCommand"]


class DebuggerCommand:
    """Base Class for Debugger commands. We pull in some helper
    functions for command from module cmdfns."""

    category = "misc"

    @staticmethod
    def setup(local_dict, category="misc", min_args=0, max_args=None, need_stack=False):
        local_dict["name"] = local_dict["__module__"].split(".")[-1]
        local_dict["category"] = category
        local_dict["min_args"] = min_args
        local_dict["max_args"] = max_args
        local_dict["need_stack"] = need_stack
        return

    def __init__(self, proc):
        """proc contains the command processor object that this
        command is invoked through.  A debugger field gives access to
        the stack frame and I/O."""
        self.proc = proc

        # Convenience class access. We don't expect that either core
        # or debugger will change over the course of the program
        # execution like errmsg(), msg(), and msg_nocr() might. (See
        # the note below on these latter 3 methods.)
        #
        self.core = proc.core
        self.debugger = proc.debugger
        self.settings = self.debugger.settings
        return

    aliases = ()
    name = "YourCommandName"

    def columnize_commands(self, commands):
        """List commands arranged in an aligned columns"""
        commands.sort()
        width = self.debugger.settings["width"]
        return columnize.columnize(commands, displaywidth=width, lineprefix="    ")

    def confirm(self, msg, default=False):
        """Convenience short-hand for self.debugger.intf[-1].confirm"""
        return self.debugger.intf[-1].confirm(msg, default)

    # Note for errmsg, msg, and msg_nocr we don't want to simply make
    # an assignment of method names like self.msg = self.debugger.intf.msg,
    # because we want to allow the interface (intf) to change
    # dynamically. That is, the value of self.debugger may change
    # in the course of the program and if we made such an method assignment
    # we wouldn't pick up that change in our self.msg
    def errmsg(self, msg, opts={}):
        """Convenience short-hand for self.debugger.intf[-1].errmsg"""
        try:
            return self.debugger.intf[-1].errmsg(msg)
        except EOFError:
            # FIXME: what do we do here?
            pass
        return None

    def msg(self, msg, opts={}):
        """Convenience short-hand for self.debugger.intf[-1].msg"""
        try:
            return self.debugger.intf[-1].msg(msg)
        except EOFError:
            # FIXME: what do we do here?
            pass
        return None

    def msg_nocr(self, msg: str, opts={}):
        """Convenience short-hand for self.debugger.intf[-1].msg_nocr"""
        try:
            return self.debugger.intf[-1].msg_nocr(msg[:1000])
        except EOFError:
            # FIXME: what do we do here?
            pass
        return None

    def rst_msg(self, text, opts={}):
        """Convert ReStructuredText and run through msg()"""
        text = Mformat.rst_text(
            text,
            "plain" == self.debugger.settings["highlight"],
            self.debugger.settings["width"],
        )
        return self.msg(text)

    def run(self, args):
        """The method that implements the debugger command.
        Help on the command comes from the docstring of this method.
        """
        raise NotImplementedError(NotImplementedMessage)

    pass

    def section(self, message, opts={}):
        if "plain" != self.settings["highlight"]:
            message = colorize("bold", message)
        else:
            message += "\n" + "-" * len(message)
            pass
        self.msg(message)


if __name__ == "__main__":
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    dd = DebuggerCommand(cp)
    dd.msg("hi")
    dd.errmsg("Don't do that")
    pass



# -*- coding: utf-8 -*-
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import threading

from trepan import exception as Mexcept

# Our local modules
from trepan.bwprocessor.command import base_cmd as Mbase_cmd


class QuitCommand(Mbase_cmd.DebuggerCommand):
    """Gently exit the debugger and debugged program.

    **Input Fields:**

       { command     => 'quit',
       }

    The program being debugged is exited raising a *DebuggerQuit* exception.

    **Output Fields: **

       { name      => 'status',
         event     => 'terminated',
         [errmsg   => <error-message-array>]
         [msg      => <message-text array>]
      }"""

    name = os.path.basename(__file__).split(".")[0]
    need_stack = False

    def nothread_quit(self, arg):
        """quit command when there's just one thread."""

        self.debugger.core.stop()
        self.debugger.core.execution_status = "Quit command"
        self.proc.response["event"] = "terminated"
        self.proc.response["name"] = "status"
        self.proc.intf[-1].msg(self.proc.response)
        raise Mexcept.DebuggerQuit

    def threaded_quit(self, arg):
        """quit command when several threads are involved."""
        self.msg("Quit for threading not fully done yet. Try 'kill'.")
        return False

    def run(self, cmd_hash):
        threading_list = threading.enumerate()
        if len(threading_list) == 1 and threading_list[0].getName() == "MainThread":
            # We just have a main thread so that's safe to quit
            return self.nothread_quit(cmd_hash)
        else:
            return self.threaded_quit(cmd_hash)
        pass


if __name__ == "__main__":
    from trepan.bwprocessor.command import mock

    d, cp = mock.dbg_setup()
    command = QuitCommand(cp)
    try:
        command.run(["quit"])
    except Mexcept.DebuggerQuit:
        print("A got 'quit' a exception. Ok, be that way - I'm going home.")
        pass

    class MyThread(threading.Thread):
        def run(self):
            command.run(["quit"])
            return

        pass

    t = MyThread()
    t.start()
    t.join()
    pass



# -*- coding: utf-8 -*-
# Copyright (C) 2013, 2015, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import types

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class MacroCommand(DebuggerCommand):
    """**macro** *macro-name* *lambda-object*

    Define *macro-name* as a debugger macro. Debugger macros get a list of
    arguments which you supply without parenthesis or commas. See below
    for an example.

    The macro (really a Python lambda) should return either a String or an
    List of Strings. The string in both cases is a debugger command.  Each
    string gets tokenized by a simple split() .  Note that macro
    processing is done right after splitting on `;;`. As a result, if the
    macro returns a string containing `;;` this will not be interpreted as
    separating debugger commands.

    If a list of strings is returned, then the first string is
    shifted from the list and executed. The remaining strings are pushed
    onto the command queue. In contrast to the first string, subsequent
    strings can contain other macros. `;;` in those strings will be
    split into separate commands.

    Here is an trivial example. The below creates a macro called `l=` which is
    the same thing as `list`:

        macro l= lambda: 'list .'

    A simple text to text substitution of one command was all that was
    needed here. But usually you will want to run several commands. So those
    have to be wrapped up into a list.

    The below creates a macro called `fin+` which issues two commands
    `finish` followed by `step`:

        macro fin+ lambda: ['finish','step']

    If you wanted to parameterize the argument of the `finish` command
    you could do that this way:

        macro fin+ lambda levels: ['finish %s' % levels ,'step']

    Invoking with:

         fin+ 3

    would expand to: `['finish 3', 'step']`

    If you were to add another parameter for `step`, the note that the
    invocation might be:

         fin+ 3 2

    rather than `fin+(3,2)` or `fin+ 3, 2`.

    See also:
    ---------

     `alias` and `info macro`.
    """

    short_help = "Define a macro"

    DebuggerCommand.setup(locals(), category="support", min_args=2)

    def run(self, args):

        cmd_name = args[1]
        cmd_argstr = self.proc.cmd_argstr[len(cmd_name) :].lstrip()
        proc_obj = None
        try:
            proc_obj = eval(cmd_argstr)
        except (SyntaxError, NameError, ValueError):
            self.errmsg("Expecting a Python lambda expression; got %s" % cmd_argstr)
            pass
        if proc_obj:
            if isinstance(proc_obj, types.FunctionType):
                self.proc.macros[cmd_name] = [proc_obj, cmd_argstr]
                self.msg('Macro "%s" defined.' % cmd_name)
            else:
                self.errmsg(
                    "Expecting a Python lambda expression; got: %s" % cmd_argstr
                )
                pass
            pass
        return

    pass


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    dbgr, cmd = Mmock.dbg_setup()
    command = MacroCommand(cmd)
    for cmdline in ["macro foo lambda a,y: x+y", "macro bad2 1+2"]:
        args = cmdline.split()
        cmd_argstr = cmdline[len(args[0]) :].lstrip()
        cmd.cmd_argstr = cmd_argstr
        command.run(args)
        pass
    print(cmd.macros)
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetPatSub(Mbase_subcmd.DebuggerSubcommand):
    """**set patsub** *from-re* *replace-string*

    Add a substitution pattern rule replacing *patsub* with
    *replace-string* anywhere it is found in source file names.  If a
    substitution rule was previously set for *from-re*, the old rule is
    replaced by the new one.

    In the following example, suppose in a docker container /mnt/project is
    the mount-point for /home/rocky/project. You are running the code
    from the docker container, but debugging this from outside of that.

    Example:
    --------

        set patsub ^/mmt/project /home/rocky/project

    """

    in_list = True
    max_args = 2
    min_abbrev = len("pats")
    min_args = 2
    short_help = "Set pattern substitution rule"

    def run(self, args):
        self.proc.add_remap_pat(args[0], args[1])

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(SetPatSub)
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2015, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.printing import print_obj


class ExamineCommand(DebuggerCommand):
    """**examine** *expr1* [*expr2* ...]

    Examine value, type and object attributes of an expression.

    In contrast to normal Python expressions, expressions should not have
    blanks which would cause shlex to see them as different tokens.

    Examples:
    ---------

        examine x+1   # ok
        examine x + 1 # not ok

    See also:
    ---------

    `pr`, `pp`, and `whatis`."""

    aliases = ("x",)
    short_help = "Examine value, type, and object attributes " "of an expression"

    DebuggerCommand.setup(
        locals(),
        category="data",
        need_stack=True,
        min_args=1,
    )

    def run(self, args):
        for arg in args[1:]:
            s = print_obj(arg, self.proc.curframe)
            self.msg(s)
            pass
        return

    pass


if __name__ == "__main__":
    import inspect
    from trepan.debugger import Trepan

    d = Trepan()
    cp = d.core.processor
    cp.curframe = inspect.currentframe()
    command = ExamineCommand(cp)
    command.run(["examine", "10"])
    me = []
    print("=" * 30)
    command.run(["examine", "me"])
    print("=" * 30)
    command.run(["examine", "DebuggerCommand"])
    pass



#!/usr/bin/env python
'Unit test for debugger command completion'

import unittest
import inspect
from xdis import IS_PYPY

from trepan import debugger as Mdebugger

line_buffer = ''


def get_line_buffer():
    return line_buffer


class TestCompletion(unittest.TestCase):

    def run_complete(self, line):
        global line_buffer
        line_buffer = line
        results = []
        i = 0
        # from trepan.api import debug; debug()
        got = self.dbgr.complete(line, i)
        while (got):
            results.append(got)
            i += 1
            got = self.dbgr.complete(line, i)
            pass
        return results

    def test_complete_identifier(self):
        from trepan.processor.command import base_cmd as mBaseCmd
        from trepan.processor import complete as mComplete
        self.dbgr = Mdebugger.Trepan()

        cmdproc = self.dbgr.core.processor
        cmdproc.curframe = inspect.currentframe()
        cmd = mBaseCmd.DebuggerCommand(cmdproc)

        self.assertEqual(mComplete.complete_id_and_builtins(cmd, 'ma'),
                         ['map', 'max'])
        self.assertEqual(mComplete.complete_identifier(cmd, 'm'),
                         ['mBaseCmd', 'mComplete'])
        return

    if not IS_PYPY:
        def test_completion(self):

            self.dbgr = Mdebugger.Trepan()

            for line, expect_completion in [
                    ['set basename ', ['off', 'on']],
                    ['where', ['where ']],  # Single alias completion
                    ['sho', ['show']],  # Simple single completion
                    ['un', ['unalias', 'undisplay']],  # Simple multiple completion
                    ['python ', []],        # Don't add anything - no more
                    ['set basename o', ['off', 'on']],
                    ['set basename of', ['off']],

                    # Multiple completion on two words
                    ['set auto', ['autoeval', 'autolist', 'autopc', 'autopython']],

                    # Completion when word is complete, without space.
                    ['show', ['show ']],

                    # Completion when word is complete with space.
                    ['info ',
                     ['args', 'break', 'builtins', 'code',
                      'display', 'files', 'frame', 'globals', 'line', "lines",
                      'locals', 'macro', "offsets", 'pc', 'program', 'return', 'signals', 'source',
                      'threads']],

                    ['help sta', ['stack', 'status']],
                    [' unalias c',  ['c', 'chdir', 'cond']],


                    # Any set style completion
                    ['set style def', ['default']],

                    # ['set auto eval ', '', ['off', 'on']],
                                              # Many 3-word completions
                    # ['set auto ', ['eval', 'irb', 'list']],
                                              # Many two-word completions
                    # ['set auto e', ['eval']],
                    # ['disas', ['disassemble']], # Another single completion
                    # ['help syn', ['syntax']],
                    # ## FIXME:
                    # ## ['help syntax co', ['command']],
                    # ['help br', ['break', 'breakpoints']],
                ]:
                got = self.run_complete(line)
                self.assertEqual(expect_completion, got,
                                 "Completion of '%s', expecting %s, got %s" %
                                 (line, expect_completion, got))
                pass

            got = self.run_complete('')
            self.assertTrue(len(got) > 30,
                            'Initial completion should return more '
                            'than 30 commands')
            got = self.run_complete('info files ')
            self.assertTrue(len(got) > 0,
                            'info files completion should return a file')
            got = self.run_complete('unalias ')
            self.assertTrue(len(got) > 0,
                            'unalias should return lots of aliases')
            return
    pass

if __name__ == '__main__':
    unittest.main()
    pass



#!/usr/bin/env python3
"Unit test for trepan.lib.eval"
import unittest

from trepan.lib import eval as Meval


class TestExtractExpression(unittest.TestCase):
    def test_extract_expression(self):
        for fragment, expect in (
            ("if condition(x):", "condition(x)"),
            ("elif is_magic(name):", "is_magic(name)"),
            ("while expression:", "expression"),
            ("for i in range(3):", "range(3)"),
            ("and x > 3", "x > 3"),
            ("or y < 3", "y < 3"),
            ("abc = 123", "123"),
            ("assert True", "True"),
            ("return return_value", "return_value"),
            ("nothing_to_be.done", "nothing_to_be.done"),
        ):
            self.assertEqual(expect, Meval.extract_expression(fragment))
            pass
        pass

    pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
'Unit test for trepan.clifns'
import inspect, os, sys, unittest

from trepan import clifns as Mclifns


class TestCLIFns(unittest.TestCase):

    def test_clifns(self):
        """Test clifns.expanduser_abs()"""
        file1 = os.path.join(os.path.curdir, "test_clifns")
        file1 = Mclifns.path_expanduser_abs(file1)
        file2 = Mclifns.path_expanduser_abs("test_clifns")
        self.assertEqual(file1, file2, "path_expanduser_abs")
        self.assertTrue(Mclifns.path_expanduser_abs("~/foo"))
        return
        
    def test_is_ok_line_for_breakpoint(self):
        filename =  __file__
        if len(filename) > 4 and filename[-4:] == '.pyc':
            filename = filename[:-1]
            pass

        # Pick up a Python code line for testing.
        # Note that this comment line relative to the line
        # we pick up is also used.
        frame = inspect.currentframe()
        lineno = frame.f_lineno
        self.assertTrue(Mclifns
                        .is_ok_line_for_breakpoint(filename,
                                                   lineno, sys.stdout.write))

        self.assertFalse(Mclifns
                         .is_ok_line_for_breakpoint(filename,
                                                    lineno-5,
                                                    sys.stdout.write))
        return

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2009-2010, 2013, 2015 Rocky Bernstein rocky@gnu.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetTrace(Mbase_subcmd.DebuggerSetBoolSubcommand):

    """**set trace** [ **on** | **off** ]

    Set event tracing.

    See also:
    ---------

    `set events`, and `show trace`."""

    in_list = True
    min_abbrev = len("trace")  # Must use at least "set trace"
    short_help = "Set event tracing"
    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(SetTrace)
    d = sub.proc.debugger
    for args in (["on"], ["off"]):
        sub.run(args)
        print(d.settings["trace"])
        pass
    pass



#!/usr/bin/env python
import unittest
from fn_helper import compare_output, strarray_setup


class TestBreak(unittest.TestCase):
    def test_break_on_function(self):
        print("test ", __file__, "break_on_function skipped")
        return

        ##############################
        # We had a bug where 'next' (no number) after
        # a hit breakpoint wouldn't 'next'. So test for that
        # in addition to the break on a function name.
        cmds = ['break foo', 'continue', 'next', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        ##############################

        def foo():
            print('foo here')
            return
        foo()
        y = 6  # NOQA
        ##############################
        d.core.stop()

        out = ['xx def foo():\n(test-break.py:18): foo',
               "-- print 'foo here'\n"]
        compare_output(self, out, d, cmds)


        # Try a break with a module name
        import os
        cmds = ['break os.path.join', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        p = os.path.join("a", "b")  # NOQA
        ##############################
        d.core.stop()
        out = ['-- p = os.path.join("a", "b")',
               'xx def join(a, *p):']
        compare_output(self, out, d, cmds)
        return

    def test_break_at_line_number(self):
        print("test", __file__, "break_at_line_number skipped")
        return
        import inspect
        curframe = inspect.currentframe()
        cmds = ['break %d' % (curframe.f_lineno+7),
                'continue', 'c']              # 1
        d = strarray_setup(cmds)              # 2
        d.core.start()                        # 3
        ##############################        # 4...
        x = 5  # NOQA
        y = 6  # NOQA
        z = 7  # NOQA
        ##############################
        d.core.stop()
        out = ["-- x = 5\nBreakpoint 1 set at line 53 of file test-break.py",
               'xx z = 7\n']
        compare_output(self, out, d, cmds)
        return
    pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
'Unit test for trepan.lib.sighandler'
import signal, unittest

from trepan.lib import sighandler as Msig


class TestLibSigHandle(unittest.TestCase):

    def test_YN(self):
        for expect, b in (('Yes', True), ('No', False)):
            self.assertEqual(expect, Msig.yes_or_no(b))
            pass
        return

    def test_canonic_signame(self):
        for expect, name_num in (('SIGTERM',  '15'),
                                 ('SIGTERM', '-15'),
                                 ('SIGTERM', 'term'),
                                 ('SIGTERM', 'sigterm'),
                                 ('SIGTERM', 'TERM'),
                                 (None, '300'),
                                 (False, 'bogus')):
            self.assertEqual(expect, Msig.canonic_signame(name_num),
                             'name_num: %s' % name_num)
            pass
        pass

    def test_lookup_signame(self):
        for expect, num in (('SIGTERM', 15), ('SIGTERM', -15),
                          (None, 300)):
            self.assertEqual(expect, Msig.lookup_signame(num), "looking up %d" % num)
            pass
        return

    def test_lookup_signum(self):
        for expect, name in ((15, 'SIGTERM'), (15, 'TERM'),
                             (15, 'term'), (None, 'nothere')):
            self.assertEqual(expect, Msig.lookup_signum(name),
                             "looking up name %s" % name)
            pass
        return

    def test_lookup_signame_signum(self):
        ignore = ['ITIMER_VIRTUAL', 'ITIMER_REAL', 'ITIMER_PROF']
        for signum in range(signal.NSIG):
            signame = Msig.lookup_signame(signum)
            if signame is not None and signame not in ignore:
                self.assertEqual(signum, Msig.lookup_signum(signame),
                                 "looking up name %s" % signame)
                # Try without the SIG prefix
                self.assertEqual(signum, Msig.lookup_signum(signame[3:]))
                pass
            pass
        return

    pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
'Unit test for trepan.inout.tcp*'
import unittest
from trepan.inout import tcpserver as Mserver, tcpclient as Mclient


class TestTCP(unittest.TestCase):
    """Tests TCPServer and TCPClient"""

    def test_client_server(self):
        client = None
        server = None
        try:
            try:
                server = Mserver.TCPServer(opts={'open': True})
            except:
                print("Skipping because of server open failure")
                return
            print("Server port is %s" % server.PORT)
            try:
                client = Mclient.TCPClient(opts={'open': True,
                                                'PORT': server.PORT})
            except IOError:
                print("Skipping because of client open failure")
                return
            for line in ['one', 'two', 'three']:
                server.writeline(line)
                self.assertEqual(line, client.read_msg().rstrip('\n'))
                pass
            for line in ['four', 'five', 'six']:
                client.writeline(line)
                self.assertEqual(line, server.read_msg().rstrip('\n'))
                pass
        finally:
            if client:
                client.close()
            if server:
                server.close()
        return

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2017-2018, 2020 Rocky Bernstein
#

from pygments.styles import STYLE_MAP

style_names = sorted(list(STYLE_MAP.keys()))

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.lib import complete as Mcomplete


class SetStyle(Mbase_subcmd.DebuggerSubcommand):
    """**set style** [*pygments-style*]

    Set the pygments style in to use in formatting text for a 256-color terminal.
    Note: if your terminal doesn't support 256 colors, you may be better off
    using `--highlight=plain` or `--highlight=dark` instead. To turn off styles
    use `set style none`.

    To list the available pygments styles inside the debugger, omit the style name.

    Examples:
    ---------

        set style            # give a list of the style names
        set style colorful   # Pygments 'colorful' style
        set style none       # Turn off style, still use highlight though

    See also:
    --------

    `show style`, `set highlight`"""

    def complete(self, prefix):
        return Mcomplete.complete_token(style_names, prefix)

    in_list = True
    min_abbrev = len("sty")
    short_help = "Set the pygments style"

    def run(self, args):
        if len(args) == 0:
            self.section("style names: ")
            self.msg(self.columnize_commands(style_names))
            return
        if args[0] == "none":
            self.debugger.settings[self.name] = None
            return

        if args[0] not in style_names:
            self.errmsg("style name '%s' not valid. valid names are: " % args[0])
            self.msg(self.columnize_commands(style_names))
            return

        self.debugger.settings[self.name] = args[0]
        return

    pass


# if __name__ == '__main__':
#     from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper
#     sub = Mhelper.demo_run(SetWidth)
#     d = sub.proc.debugger
#     sub.run(['emacs'])
#     print(d.settings['emacs'])
#     pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetAutoEval(Mbase_subcmd.DebuggerSetBoolSubcommand):
    """**set** **autoeval** [**on**|**off**]

    Evaluate unrecognized debugger commands.

    Often inside the debugger, one would like to be able to run arbitrary
    Python commands without having to preface Python expressions with `print` or
    `eval`. Setting *autoeval* on will cause unrecognized debugger
    commands to be *eval*'d as a Python expression.

    Note that if this is set, on error the message shown on type a bad
    debugger command changes from:

      Undefined command: "fdafds". Try "help".

    to something more Python-eval-specific such as:

      NameError: name 'fdafds' is not defined

    One other thing that trips people up is when setting autoeval is that
    there are some short debugger commands that sometimes one wants to use
    as a variable, such as in an assignment statement. For example:

      s = 5

    which produces when *autoeval* is on:

      Command 'step' can take at most 1 argument(s); got 2.

    because by default, `s` is an alias for the debugger `step`
    command. It is possible to remove that alias if this causes constant
    problem.

    Another possibility is to go into a real Python shell via the `python`
    or `ipython` commands."""

    short_help = "Evaluate unrecognized debugger commands."
    in_list = True
    min_abbrev = len("autoe")
    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(SetAutoEval)
    pass



#!/usr/bin/env python3
import unittest
import tracer
from fn_helper import strarray_setup, compare_output


class TestStep(unittest.TestCase):
    print("test ", __file__, "skipped")

    def test_step_same_level(self):
        return

        # See that we can step with parameter which is the same as 'step 1'
        cmds = ['step', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5  # NOQA
        y = 6  # NOQA
        ##############################
        d.core.stop()
        out = ['-- x = 5',
               '-- y = 6\n']
        compare_output(self, out, d, cmds)
        return

    @unittest.skip("Need to fix")
    def test_step_computed_valued(self):
        # See that we can step with a computed count value
        cmds = ['step 5-3', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5
        y = 6
        z = 7
        ##############################
        d.core.stop(options={'remove': True})
        out = ['-- x = 5',
               '-- z = 7']
        compare_output(self, out, d, cmds)

        # Test step>
        cmds = ['step>', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5

        def foo():
            return
        y = 6  # NOQA
        foo()
        ##############################
        d.core.stop(options={'remove': True})
        out = ['-- x = 5',
               '-> def foo():']
        compare_output(self, out, d, cmds)

        # # Test step!
        # cmds = ['step!', 'continue']
        # d = strarray_setup(cmds)
        # d.core.start()
        # ##############################
        # x = 5
        # try:
        #     y = 2
        #     z = 1/0
        # except:
        #     pass
        # ##############################
        # d.core.stop(options={'remove': True})
        # out = ['-- x = 5',
        #        '!! z = 1/0\n']
        # compare_output(self, out, d, cmds)

        # Test "step" with sets of events. Part 1
        cmds = ['step call exception',
                'step call exception', 'continue']
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5  # NOQA
        try:
            def foo1():
                y = 2  # NOQA
                raise Exception
                return
            foo1()
        except:
            pass
        z = 1  # NOQA
        # ##############################
        # d.core.stop(options={'remove': True})
        # out = ['-- x = 5',
        #        '-> def foo1():',
        #        '!! raise Exception']
        # compare_output(self, out, d, cmds)

        # # Test "step" will sets of events. Part 2
        # cmds = ['step call exception 1+0',
        #         'step call exception 1', 'continue']
        # d = strarray_setup(cmds)
        # d.core.start()
        # ##############################
        # x = 5
        # try:
        #     def foo2():
        #         y = 2
        #         raise Exception
        #         return
        #     foo2()
        # except:
        #     pass
        # z = 1
        # ##############################
        # d.core.stop(options={'remove': True})
        # out = ['-- x = 5',
        #        '-> def foo2():',
        #        '!! raise Exception']
        # compare_output(self, out, d, cmds)

        return

    @unittest.skip("Need to fix")
    def test_step_between_fn(self):

        # Step into and out of a function
        def sqr(x):
            return x * x
        for cmds, out, eventset in (
            (['step', 'step', 'continue'],
             ['-- x = sqr(4)   # NOQA',
              '-- return x * x',
              '-- y = 5  # NOQA'],
             frozenset(('line',))),
            (['step', 'step', 'step', 'step', 'continue'],
             ['-- d.core.start()',
              '-- x = sqr(4)   # NOQA',
               '-> def sqr(x):',
               '-- return x * x',
               '<- return x * x'],
             tracer.ALL_EVENTS), ):
            d = strarray_setup(cmds)
            d.settings['events'] = eventset
            d.core.start()
            ##############################
            x = sqr(4)   # NOQA
            y = 5  # NOQA
            ##############################
            d.core.stop(options={'remove': True})
            compare_output(self, out, d, cmds)
            pass
        return

    @unittest.skip("Need to fix")
    def test_step_in_exception(self):
        return
        def boom(x):
            y = 0/x  # NOQA
            return

        def bad(x):
            boom(x)
            return x * x
        cmds = ['step', 'step', 'step', 'step', 'step', 'step',
                'step', 'step', 'step', 'step', 'continue']
        d = strarray_setup(cmds)
        try:
            d.core.start()
            x = bad(0)
            self.assertTrue(False, 'should have raised an exception')
        except ZeroDivisionError:
            self.assertTrue(True, 'Got the exception')
        finally:
            d.core.stop(options={'remove': True})
            pass

        out = ['-- x = bad(0)  # NOQA', # line event
               '-> def bad(x):',        # call event
               '-- boom(x)',            # line event
               '-> def boom(x):',       # call event
               '-- y = 0/x  # NOQA',    # line event
               '!! y = 0/x  # NOQA',    # exception event
               '<- y = 0/x  # NOQA',    # return event
               '!! boom(x)',            # exception event
               '<- boom(x)',            # return event
               '!! x = bad(0)  # NOQA', # exception event
               '-- except ZeroDivisionError:']
        compare_output(self, out, d, cmds)
        return

    pass

if __name__ == '__main__':
    unittest.main()
    pass



"""
Twisted integration
-------------------

This module provides a very simple way to integrate your tests with the
Twisted_ event loop.

You must import this module *before* importing anything from Twisted itself!

Example::

  from nose.twistedtools import reactor, deferred
  
  @deferred()
  def test_resolve():
      return reactor.resolve("www.python.org")

Or, more realistically::

  @deferred(timeout=5.0)
  def test_resolve():
      d = reactor.resolve("www.python.org")
      def check_ip(ip):
          assert ip == "67.15.36.43"
      d.addCallback(check_ip)
      return d

.. _Twisted: http://twistedmatrix.com/trac/
"""

import sys
from queue import Queue, Empty
from nose.tools import make_decorator, TimeExpired

__all__ = [
    'threaded_reactor', 'reactor', 'deferred', 'TimeExpired',
    'stop_reactor'
]

_twisted_thread = None

def threaded_reactor():
    """
    Start the Twisted reactor in a separate thread, if not already done.
    Returns the reactor.
    The thread will automatically be destroyed when all the tests are done.
    """
    global _twisted_thread
    try:
        from twisted.internet import reactor
    except ImportError:
        return None, None
    if not _twisted_thread:
        from twisted.python import threadable
        from threading import Thread
        _twisted_thread = Thread(target=lambda: reactor.run( \
                installSignalHandlers=False))
        _twisted_thread.setDaemon(True)
        _twisted_thread.start()
    return reactor, _twisted_thread

# Export global reactor variable, as Twisted does
reactor, reactor_thread = threaded_reactor()


def stop_reactor():
    """Stop the reactor and join the reactor thread until it stops.
    Call this function in teardown at the module or package level to
    reset the twisted system after your tests. You *must* do this if
    you mix tests using these tools and tests using twisted.trial.
    """
    global _twisted_thread

    def stop_reactor():
        '''Helper for calling stop from withing the thread.'''
        reactor.stop()

    reactor.callFromThread(stop_reactor)
    reactor_thread.join()
    for p in reactor.getDelayedCalls():
        if p.active():
            p.cancel()
    _twisted_thread = None


def deferred(timeout=None):
    """
    By wrapping a test function with this decorator, you can return a
    twisted Deferred and the test will wait for the deferred to be triggered.
    The whole test function will run inside the Twisted event loop.

    The optional timeout parameter specifies the maximum duration of the test.
    The difference with timed() is that timed() will still wait for the test
    to end, while deferred() will stop the test when its timeout has expired.
    The latter is more desireable when dealing with network tests, because
    the result may actually never arrive.

    If the callback is triggered, the test has passed.
    If the errback is triggered or the timeout expires, the test has failed.

    Example::
    
        @deferred(timeout=5.0)
        def test_resolve():
            return reactor.resolve("www.python.org")

    Attention! If you combine this decorator with other decorators (like
    "raises"), deferred() must be called *first*!

    In other words, this is good::
        
        @raises(DNSLookupError)
        @deferred()
        def test_error():
            return reactor.resolve("xxxjhjhj.biz")

    and this is bad::
        
        @deferred()
        @raises(DNSLookupError)
        def test_error():
            return reactor.resolve("xxxjhjhj.biz")
    """
    reactor, reactor_thread = threaded_reactor()
    if reactor is None:
        raise ImportError("twisted is not available or could not be imported")
    # Check for common syntax mistake
    # (otherwise, tests can be silently ignored
    # if one writes "@deferred" instead of "@deferred()")
    try:
        timeout is None or timeout + 0
    except TypeError:
        raise TypeError("'timeout' argument must be a number or None")

    def decorate(func):
        def wrapper(*args, **kargs):
            q = Queue()
            def callback(value):
                q.put(None)
            def errback(failure):
                # Retrieve and save full exception info
                try:
                    failure.raiseException()
                except:
                    q.put(sys.exc_info())
            def g():
                try:
                    d = func(*args, **kargs)
                    try:
                        d.addCallbacks(callback, errback)
                    # Check for a common mistake and display a nice error
                    # message
                    except AttributeError:
                        raise TypeError("you must return a twisted Deferred "
                                        "from your test case!")
                # Catch exceptions raised in the test body (from the
                # Twisted thread)
                except:
                    q.put(sys.exc_info())
            reactor.callFromThread(g)
            try:
                error = q.get(timeout=timeout)
            except Empty:
                raise TimeExpired("timeout expired before end of test (%f s.)"
                                  % timeout)
            # Re-raise all exceptions
            if error is not None:
                exc_type, exc_value, tb = error
                raise exc_type(exc_value).with_traceback(tb)
        wrapper = make_decorator(func)(wrapper)
        return wrapper
    return decorate




#!/usr/bin/env python3
# -*- coding: iso-8859-1 -*-
#   Copyright (C) 2008-2010, 2013-2018, 2020-2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The command-line interface to the debugger.
"""
import os
import os.path as osp
import sys
import tempfile

import pyficache

package = "trepan"

# Our local modules
import trepan.client as Mclient
import trepan.clifns as Mclifns
import trepan.debugger as Mdebugger
import trepan.exception as Mexcept
import trepan.interfaces.server as Mserver
import trepan.lib.file as Mfile
import trepan.misc as Mmisc
import trepan.options as Moptions

# The name of the debugger we are currently going by.
__title__ = package

from trepan.version import __version__


def main(dbg=None, sys_argv=list(sys.argv)):
    """Routine which gets run if we were invoked directly"""
    global __title__

    # Save the original just for use in the restart that works via exec.
    orig_sys_argv = list(sys_argv)
    opts, dbg_opts, sys_argv = Moptions.process_options(
        __title__, __version__, sys_argv
    )

    if opts.server is not None:
        if opts.server == "tcp":
            connection_opts = {"IO": "TCP", "PORT": opts.port}
        else:
            connection_opts = {"IO": "FIFO"}
        intf = Mserver.ServerInterface(connection_opts=connection_opts)
        dbg_opts["interface"] = intf
        if "FIFO" == intf.server_type:
            print("Starting FIFO server for process %s." % os.getpid())
        elif "TCP" == intf.server_type:
            print("Starting TCP server listening on port %s." % intf.inout.PORT)
            pass
    elif opts.client:
        Mclient.run(opts, sys_argv)
        return

    dbg_opts["orig_sys_argv"] = orig_sys_argv

    if dbg is None:
        dbg = Mdebugger.Trepan(dbg_opts)
        dbg.core.add_ignore(main)
        pass
    Moptions._postprocess_options(dbg, opts)

    # process_options has munged sys.argv to remove any options that
    # options that belong to this debugger. The original options to
    # invoke the debugger and script are in global sys_argv

    if len(sys_argv) == 0:
        # No program given to debug. Set to go into a command loop
        # anyway
        mainpyfile = None
    else:
        mainpyfile = sys_argv[0]  # Get script filename.
        if not osp.isfile(mainpyfile):
            mainpyfile = Mclifns.whence_file(mainpyfile)
            is_readable = Mfile.readable(mainpyfile)
            if is_readable is None:
                print(
                    "%s: Python script file '%s' does not exist"
                    % (
                        __title__,
                        mainpyfile,
                    )
                )
                sys.exit(1)
            elif not is_readable:
                print(
                    "%s: Can't read Python script file '%s'"
                    % (
                        __title__,
                        mainpyfile,
                    )
                )
                sys.exit(1)
                return

        if Mfile.is_compiled_py(mainpyfile):
            try:
                from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE, load_module
                from xdis.version_info import version_tuple_to_str

                (
                    python_version,
                    timestamp,
                    magic_int,
                    co,
                    is_pypy,
                    source_size,
                    sip_hash,
                ) = load_module(mainpyfile, code_objects=None, fast_load=False)
                if is_pypy != IS_PYPY:
                    print("Bytecode is pypy %s, but we are %s." % (is_pypy, IS_PYPY))
                    print("For a cross-version debugger, use trepan-xpy with x-python.")
                    sys.exit(2)
                if python_version[:2] != PYTHON_VERSION_TRIPLE[:2]:
                    print(
                        "Bytecode is for version %s but we are version %s."
                        % (python_version, version_tuple_to_str())
                    )
                    print("For a cross-version debugger, use trepan-xpy with x-python.")
                    sys.exit(2)

                py_file = co.co_filename
                if osp.isabs(py_file):
                    try_file = py_file
                else:
                    mainpydir = osp.dirname(mainpyfile)
                    tag = sys.implementation.cache_tag
                    dirnames = (
                        [osp.join(mainpydir, tag), mainpydir]
                        + os.environ["PATH"].split(osp.pathsep)
                        + ["."]
                    )
                    try_file = Mclifns.whence_file(py_file, dirnames)

                if osp.isfile(try_file):
                    mainpyfile = try_file
                    pass
                else:
                    # Move onto the except branch
                    raise IOError(
                        "Python file name embedded in code %s not found" % try_file
                    )
            except IOError:
                decompiler = "uncompyle6"
                try:
                    if (3, 7) <= PYTHON_VERSION_TRIPLE <= (3, 8):
                        from uncompyle6 import decompile_file
                    else:
                        from decompyle3 import decompile_file

                        decompiler = "decompyle3"
                except ImportError:
                    print(
                        "%s: Compiled python file '%s', but %s not found"
                        % (__title__, mainpyfile, decompiler),
                        file=sys.stderr,
                    )
                    sys.exit(1)
                    return

                short_name = osp.basename(mainpyfile).strip(".pyc")
                fd = tempfile.NamedTemporaryFile(
                    suffix=".py",
                    prefix=short_name + "_",
                    delete=False,
                    dir=dbg.settings["tempdir"],
                )
                old_write = fd.file.write

                def write_wrapper(*args, **kwargs):
                    if isinstance(args[0], str):
                        new_args = list(args)
                        new_args[0] = args[0].encode("utf-8")
                        old_write(*new_args, **kwargs)
                    else:
                        old_write(*args, **kwargs)

                fd.file.write = write_wrapper

                # from io import StringIO
                # linemap_io = StringIO()
                try:
                    decompile_file(mainpyfile, fd.file, mapstream=fd)
                except Exception:
                    print(
                        "%s: error decompiling '%s'" % (__title__, mainpyfile),
                        file=sys.stderr,
                    )
                    sys.exit(1)
                    return

                # # Get the line associations between the original and
                # # decompiled program
                # mapline = linemap_io.getvalue()
                # fd.write(mapline + "\n\n")
                # linemap = eval(mapline[3:])
                mainpyfile = fd.name
                fd.close()

                # Since we are actually running the recreated source,
                # there is little no need to remap line numbers.
                # The mapping is given at the end of the file.
                # However we should consider adding this information
                # and original file name.

                print(
                    "%s: couldn't find Python source so we recreated it at '%s'"
                    % (__title__, mainpyfile),
                    file=sys.stderr,
                )

                pass

        # If mainpyfile is an optimized Python script try to find and
        # use non-optimized alternative.
        mainpyfile_noopt = pyficache.resolve_name_to_path(mainpyfile)
        if mainpyfile != mainpyfile_noopt and Mfile.readable(mainpyfile_noopt):
            print("%s: Compiled Python script given and we can't use that." % __title__)
            print(
                "%s: Substituting non-compiled name: %s"
                % (
                    __title__,
                    mainpyfile_noopt,
                )
            )
            mainpyfile = mainpyfile_noopt
            pass

        # Replace trepan's dir with script's dir in front of
        # module search path.
        sys.path[0] = dbg.main_dirname = osp.dirname(mainpyfile)

    # XXX If a signal has been received we continue in the loop, otherwise
    # the loop exits for some reason.
    dbg.sig_received = False

    # if not mainpyfile:
    #     print('For now, you need to specify a Python script name!')
    #     sys.exit(2)
    #     pass

    while True:
        # Run the debugged script over and over again until we get it
        # right.

        try:
            if dbg.program_sys_argv and mainpyfile:
                normal_termination = dbg.run_script(mainpyfile)
                if not normal_termination:
                    break
            else:
                dbg.core.execution_status = "No program"
                dbg.core.processor.process_commands()
                pass

            dbg.core.execution_status = "Terminated"
            dbg.intf[-1].msg("The program finished - quit or restart")
            dbg.core.processor.process_commands()
        except Mexcept.DebuggerQuit:
            break
        except Mexcept.DebuggerRestart:
            dbg.core.execution_status = "Restart requested"
            if dbg.program_sys_argv:
                sys.argv = list(dbg.program_sys_argv)
                part1 = "Restarting %s with arguments:" % dbg.core.filename(mainpyfile)
                args = " ".join(dbg.program_sys_argv[1:])
                dbg.intf[-1].msg(
                    Mmisc.wrapped_lines(part1, args, dbg.settings["width"])
                )
            else:
                break
        except SystemExit:
            # In most cases SystemExit does not warrant a post-mortem session.
            break
        pass

    # Restore old sys.argv
    sys.argv = orig_sys_argv
    return


# When invoked as main program, invoke the debugger on a script
if __name__ == "__main__":
    main()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.lib import complete as Mcomplete
from trepan.processor import frame as Mframe


class InfoFrame(Mbase_subcmd.DebuggerSubcommand):
    """**info frame** [-v] [ *frame-number* | *frame-object* ]

    Show the detailed information for *frame-number* or the current frame if
    *frame-number* is not specified. You can also give a frame object instead of
    a frame number

    Specific information includes:

    * the frame number (if not an object)

    * the source-code line number that this frame is stopped in

    * the last instruction executed; -1 if the program are before the first
    instruction

    * a function that tracing this frame or `None`

    * Whether the frame is in restricted execution

    If `-v` is given we show builtin and global names the frame sees.

    See also:
    ---------

    `info locals`, `info globals`, `info args`"""

    min_abbrev = 2
    max_args = 2
    need_stack = True
    short_help = """Show detailed info about the current frame"""

    def complete(self, prefix):
        proc_obj = self.proc
        low, high = Mframe.frame_low_high(proc_obj, None)
        ary = [str(low + i) for i in range(high - low + 1)]
        # FIXME: add in Thread names
        return Mcomplete.complete_token(ary, prefix)

    def run(self, args):

        # FIXME: should DRY this with code.py
        proc = self.proc
        frame = proc.curframe
        if not frame:
            self.errmsg("No frame selected.")
            return False

        show_lists = False
        if len(args) >= 1 and args[0] == "-v":
            args.pop(0)
            show_lists = True

        frame_num = None
        if len(args) == 1:
            try:
                frame_num = int(args[0])
                i_stack = len(proc.stack)
                if frame_num < -i_stack or frame_num > i_stack - 1:
                    self.errmsg(
                        (
                            "a frame number number has to be in the range %d to %d."
                            + " Got: %d (%s)."
                        )
                        % (-i_stack, i_stack - 1, frame_num, args[0])
                    )
                    return False
                frame = proc.stack[frame_num][0]
            except Exception:
                try:
                    frame = eval(args[0], frame.f_globals, frame.f_locals)
                except Exception:
                    self.errmsg(
                        "%s is not a evaluable as a frame object or frame number."
                        % args[0]
                    )
                    return False
                if not inspect.isframe(frame):
                    self.errmsg("%s is not a frame object" % frame)
                pass
        else:
            frame_num = proc.curindex

        mess = (
            "Frame %d" % Mframe.frame_num(proc, frame_num)
            if frame_num is not None and proc.stack is not None
            else "Frame Info"
        )
        self.section(mess)
        if hasattr(frame, "f_restricted"):
            self.msg("  restricted execution: %s" % frame.f_restricted)
        self.msg("  current line number: %d" % frame.f_lineno)
        self.msg("  last instruction: %d" % frame.f_lasti)
        self.msg("  code: %s" % frame.f_code)
        self.msg("  previous frame: %s" % frame.f_back)
        self.msg("  tracing function: %s" % frame.f_trace)

        if show_lists:
            for name, field in [
                ("Globals", "f_globals"),
                ("Builtins", "f_builtins"),
            ]:
                vals = getattr(frame, field).keys()
                if vals:
                    self.section(name)
                    m = self.columnize_commands(vals)
                    self.msg_nocr(m)

        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock, info as Minfo

    d, cp = mock.dbg_setup()
    cp.setup()
    i = Minfo.InfoCommand(cp)

    sub = InfoFrame(i)
    cp.curframe = inspect.currentframe()
    sub.run([])



#!/usr/bin/env python3
'Unit test for trepan.lib.display'

import inspect, unittest

from trepan.lib import display as Mdisplay


class TestLibDisplay(unittest.TestCase):

    def test_display(self):

        mgr = Mdisplay.DisplayMgr()
        self.assertEqual(mgr.list, [], "Initial list empty")
        x = 1  # NOQA
        frame = inspect.currentframe()
        disp = mgr.add(frame, 'x > 1')
        self.assertEqual(disp.__class__, Mdisplay.Display,
                         "mgr.add should return display")
        self.assertEqual(len(mgr.list), 1, "display list with one item")
        disp = mgr.add(frame, 'x')
        self.assertEqual(disp.__class__, Mdisplay.Display,
                         "mgr.add should return another display")
        self.assertEqual(len(mgr.list), 2, "display list with two items")
        self.assertEqual(mgr.delete_index(1), True, "return True on ok delete")
        self.assertEqual(mgr.delete_index(1), False,
                         "return False on no delete")
        self.assertEqual(len(mgr.list), 1, "display list again with one item")
        return

    pass

if __name__ == '__main__':
    unittest.main()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013-2014, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


class DebuggerQuit(Exception):
    """An exception to signal a graceful termination of the program"""

    pass


class DebuggerRestart(Exception):
    """An exception to signal a (soft) program restart.
    You can pass in an array containing the arguments to restart, should
    we have to issue an execv-style restart.
    """

    def __init__(self, sys_argv=[]):
        self.sys_argv = sys_argv
        return

    pass


if __name__ == "__main__":
    try:
        raise DebuggerRestart(["a", "b"])
    except DebuggerRestart:
        import sys

        print(sys.exc_info()[1].sys_argv)
        pass
    pass
